"""
Tagulous test: Tag Trees

Modules tested:
    tagulous.models.tree
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

from tests.lib import *
import tagulous.settings as tagulous_settings


class TagTreeTestManager(TagTestManager):
    """
    Manage tag tree tests
    """
    def assertTreeTag(
        self, tag, name=None, label=None, slug=None, path=None,
        parent=None, count=None, protected=None, level=None,
    ):
        "Check tag attributes match those specified"
        if name is not None:
            self.assertEqual(tag.name, name)
        if label is not None:
            self.assertEqual(tag.label, label)
        if slug is not None:
            self.assertEqual(tag.slug, slug)
        if path is not None:
            self.assertEqual(tag.path, path)

        if parent is not None:
            self.assertEqual(tag.parent, parent)
        if count is not None:
            self.assertEqual(tag.count, count)
        if protected is not None:
            self.assertEqual(tag.protected, protected)
        if level is not None:
            self.assertEqual(tag.level, level)


###############################################################################
####### TagTreeModel basics
###############################################################################

class TagTreeModelTest(TagTreeTestManager, TestCase):
    """
    Test TagTreeModel basics - that the models are created correctly, and tag
    instances can be created directly with parents and fields correct.

    These tests refer to levels; the higher the level, the deeper in the tree,
    eg: level1/level2/level3
    """
    manage_models = [
        test_models.TreeTest,
    ]

    def setUpExtra(self):
        self.test_model = test_models.TreeTest
        self.singletag_field = test_models.TreeTest.singletag
        self.singletag_model = test_models.TreeTest.singletag.tag_model
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

    def test_label_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(self.tag_model._meta.get_field('label').max_length,
                         tagulous_settings.LABEL_MAX_LENGTH)

    def test_singletag_model(self):
        "Check tag tree model is used as a base when SingleTagField tree=True"
        self.assertTrue(issubclass(
            self.singletag_model, tag_models.TagTreeModel
        ))

    def test_tag_model(self):
        "Check tag tree model is used as a base when TagField tree=True"
        self.assertTrue(issubclass(
            self.tag_model, tag_models.TagTreeModel
        ))

    def test_level_1(self):
        "Check level 1 node created correctly"
        t1 = self.tag_model.objects.create(name='One')
        self.assertTreeTag(
            t1, name='One', label='One', slug='one', path='one', level=1,
        )

    def test_level_2_existing_l1(self):
        "Check level 2 node created with existing level 1"
        t1 = self.tag_model.objects.create(name='One')
        t2 = self.tag_model.objects.create(name='One/Two')
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )

    def test_level_2_missing_l1(self):
        "Check level 2 node creates missing level 1"
        t2 = self.tag_model.objects.create(name='One/Two')
        t1 = self.tag_model.objects.get(name='One')
        self.assertTreeTag(
            t1, name='One', label='One', slug='one', path='one', level=1,
        )
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )

    def test_level_3_existing_l1_l2(self):
        "Check level 3 node created with existing level 1 and 2"
        t1 = self.tag_model.objects.create(name='One')
        t2 = self.tag_model.objects.create(name='One/Two')
        t3 = self.tag_model.objects.create(name='One/Two/Three')
        self.assertTreeTag(
            t3, name='One/Two/Three', label='Three',
            slug='three', path='one/two/three',
            parent=t2, level=3,
        )

    def test_level_3_existing_l1_missing_l2(self):
        "Check level 3 node created with existing level 1 but missing level 2"
        t1 = self.tag_model.objects.create(name='One')
        t3 = self.tag_model.objects.create(name='One/Two/Three')
        t2 = self.tag_model.objects.get(name='One/Two')
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', label='Three',
            slug='three', path='one/two/three',
            parent=t2, level=3,
        )

    def test_level_3_missing_l1_l2(self):
        "Check level 3 node created with missing level 1 and 2"
        t3 = self.tag_model.objects.create(name='One/Two/Three')
        t1 = self.tag_model.objects.get(name='One')
        t2 = self.tag_model.objects.get(name='One/Two')
        self.assertTreeTag(
            t1, name='One', label='One', slug='one', path='one', level=1,
        )
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', label='Three',
            slug='three', path='one/two/three',
            parent=t2, level=3,
        )

    def test_level_1_renames_l2_l3(self):
        "Check renaming level 1 node renames l2 and l3"
        t3 = self.tag_model.objects.create(name='One/Two/Three')
        t1 = self.tag_model.objects.get(name='One')
        self.assertTagModel(self.tag_model, {
            'One': 0,
            'One/Two': 0,
            'One/Two/Three': 0,
        })
        t1.name = 'Uno'
        t1.slug = None # Force slug to be recreated
        t1.save()

        # Reload
        self.assertTagModel(self.tag_model, {
            'Uno': 0,
            'Uno/Two': 0,
            'Uno/Two/Three': 0,
        })
        t1 = self.tag_model.objects.get(name='Uno')
        t2 = self.tag_model.objects.get(name='Uno/Two')
        t3 = self.tag_model.objects.get(name='Uno/Two/Three')

        self.assertTreeTag(
            t1, name='Uno', label='Uno', slug='uno', path='uno', level=1,
        )
        self.assertTreeTag(
            t2, name='Uno/Two', label='Two', slug='two', path='uno/two',
            parent=t1, level=2,
        )
        self.assertTreeTag(
            t3, name='Uno/Two/Three', label='Three',
            slug='three', path='uno/two/three',
            parent=t2, level=3,
        )

    def test_rebuild(self):
        "Check rebuild updates all slugs"
        # Break slugs
        t3 = self.tag_model.objects.create(name='One/Two/Three')
        for tag in self.tag_model.objects.all():
            tag.slug = six.text_type(tag.pk)
            tag.save()

        # Load
        t1 = self.tag_model.objects.get(name='One')
        t2 = self.tag_model.objects.get(name='One/Two')
        t3 = self.tag_model.objects.get(name='One/Two/Three')
        t1_slug = six.text_type(t1.pk)
        t2_slug = six.text_type(t2.pk)
        t3_slug = six.text_type(t3.pk)
        self.assertRegex(t1_slug, r'^\d+$')
        self.assertRegex(t2_slug, r'^\d+$')
        self.assertRegex(t3_slug, r'^\d+$')
        self.assertTreeTag(t1, name='One', slug=t1_slug, path=t1_slug)
        self.assertTreeTag(
            t2, name='One/Two', slug=t2_slug,
            path='{}/{}'.format(t1_slug, t2_slug), parent=t1,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', slug=t3_slug,
            path='{}/{}/{}'.format(t1_slug, t2_slug, t3_slug), parent=t2
        )

        # Rebuild and re-test
        self.tag_model.objects.rebuild()
        t1 = self.tag_model.objects.get(name='One')
        t2 = self.tag_model.objects.get(name='One/Two')
        t3 = self.tag_model.objects.get(name='One/Two/Three')
        self.assertTreeTag(t1, name='One', slug='one', path='one')
        self.assertTreeTag(
            t2, name='One/Two', slug='two', path='one/two', parent=t1,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', slug='three', path='one/two/three',
            parent=t2,
        )


###############################################################################
####### TagTreeModel tree merging
###############################################################################

class TagTreeModelMergeTest(TagTreeTestManager, TestCase):
    """
    Test merging TagTreeModel nodes
    """
    manage_models = [
        test_models.TreeTest,
    ]

    def setUpExtra(self):
        self.test_model = test_models.TreeTest
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        # Normal Animal leaf nodes
        self.test_model.objects.create(name='cat1', tags='Animal/Mammal/Cat')
        self.test_model.objects.create(name='dog1', tags='Animal/Mammal/Dog')
        self.test_model.objects.create(name='bee1', tags='Animal/Insect/Bee')

        # Pet tree has Cat leaf and extra L2 not in Animal
        self.test_model.objects.create(name='cat2', tags='Pet/Mammal/Cat')
        self.test_model.objects.create(name='robin1', tags='Pet/Bird/Robin')

        # Food tree has Cat leaf and extra L3 not in Animal
        self.test_model.objects.create(name='cat3', tags='Food/Mammal/Cat')
        self.test_model.objects.create(name='pig1', tags='Food/Mammal/Pig')

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(self.tag_model, {
            'Animal':               0,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        0,
            'Animal/Mammal/Cat':    1,
            'Animal/Mammal/Dog':    1,
            'Pet':                  0,
            'Pet/Mammal':           0,
            'Pet/Mammal/Cat':       1,
            'Pet/Bird':             0,
            'Pet/Bird/Robin':       1,
            'Food':                 0,
            'Food/Mammal':          0,
            'Food/Mammal/Cat':      1,
            'Food/Mammal/Pig':      1,
        })

    def test_merge_l1(self):
        """
        Test that merging an l1 without merging children reassigns tagged
        items, but leaves self and children
        """
        # Tag something with 'Pet'
        t1 = self.test_model.objects.create(name='pet1', tags='Pet')
        self.assertEqual(
            self.tag_model.objects.get(name='Pet').count, 1,
        )

        # Now merge
        self.tag_model.objects.get(name='Animal').merge_tags(
            'Pet, Food', children=False,
        )
        self.assertTagModel(self.tag_model, {
            'Animal':               1,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        0,
            'Animal/Mammal/Cat':    1,
            'Animal/Mammal/Dog':    1,
            'Pet':                  0,
            'Pet/Mammal':           0,
            'Pet/Mammal/Cat':       1,
            'Pet/Bird':             0,
            'Pet/Bird/Robin':       1,
            'Food':                 0,
            'Food/Mammal':          0,
            'Food/Mammal/Cat':      1,
            'Food/Mammal/Pig':      1,
        })
        self.assertInstanceEqual(t1, name='pet1', tags='Animal')

    def test_merge_l1_with_children(self):
        """
        Test that merging an l1 and its children merges the full subtrees and
        cleans empty parents
        """
        # Now merge
        self.tag_model.objects.get(name='Animal').merge_tags(
            'Pet, Food', children=True,
        )
        self.assertTagModel(self.tag_model, {
            'Animal':               0,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        0,
            'Animal/Mammal/Cat':    3,
            'Animal/Mammal/Dog':    1,
            'Animal/Mammal/Pig':    1,
            'Animal/Bird':          0,
            'Animal/Bird/Robin':    1,
        })

    def test_merge_l2(self):
        """
        Test that merging an l2 without merging children reassigns tagged
        items, but leaves self and children
        """
        # Tag something with 'Pet/Mammal'
        t1 = self.test_model.objects.create(name='mammal1', tags='Pet/Mammal')
        self.assertEqual(
            self.tag_model.objects.get(name='Pet/Mammal').count, 1,
        )

        # Now merge
        self.tag_model.objects.get(name='Animal/Mammal').merge_tags(
            'Pet/Mammal, Food/Mammal', children=False,
        )
        self.assertTagModel(self.tag_model, {
            'Animal':               0,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        1,
            'Animal/Mammal/Cat':    1,
            'Animal/Mammal/Dog':    1,
            'Pet':                  0,
            'Pet/Mammal':           0,
            'Pet/Mammal/Cat':       1,
            'Pet/Bird':             0,
            'Pet/Bird/Robin':       1,
            'Food':                 0,
            'Food/Mammal':          0,
            'Food/Mammal/Cat':      1,
            'Food/Mammal/Pig':      1,
        })
        self.assertInstanceEqual(t1, name='mammal1', tags='Animal/Mammal')

    def test_merge_l2_with_children(self):
        """
        Test that merging an l2 and its children merges the full subtrees and
        cleans empty parents
        """
        # Now merge
        self.tag_model.objects.get(name='Animal/Mammal').merge_tags(
            'Pet/Mammal, Food/Mammal', children=True,
        )
        self.assertTagModel(self.tag_model, {
            'Animal':               0,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        0,
            'Animal/Mammal/Cat':    3,
            'Animal/Mammal/Dog':    1,
            'Animal/Mammal/Pig':    1,
            'Pet':                  0,
            'Pet/Bird':             0,
            'Pet/Bird/Robin':       1,
        })

    def test_merge_l3(self):
        "Test merging an l3 cleans empty parents"
        # No need to test children=True - there are no children
        self.tag_model.objects.get(name='Animal/Mammal/Cat').merge_tags(
            'Pet/Mammal/Cat, Food/Mammal/Cat', children=False,
        )
        self.assertTagModel(self.tag_model, {
            'Animal':               0,
            'Animal/Insect':        0,
            'Animal/Insect/Bee':    1,
            'Animal/Mammal':        0,
            'Animal/Mammal/Cat':    3,
            'Animal/Mammal/Dog':    1,
            'Pet':                  0,
            # Pet/Mammal now empty
            'Pet/Bird':             0,
            'Pet/Bird/Robin':       1,
            'Food':                 0,
            'Food/Mammal':          0,
            'Food/Mammal/Pig':      1,
        })


###############################################################################
####### Custom TagTreeModel basics
###############################################################################

class TagTreeModelCustomTest(TagTreeTestManager, TransactionTestCase):
    """
    Test TagTreeModel basics when using a custom model

    Use a TransactionTestCase so the apps will be reset
    """
    manage_models = [
        test_models.CustomTreeTest,
    ]

    def setUpExtra(self):
        self.test_model = test_models.CustomTreeTest
        self.tag_model = test_models.CustomTagTree
        self.singletag_field = test_models.CustomTreeTest.singletag
        self.tag_field = test_models.CustomTreeTest.tags

    def test_tag_model(self):
        "Check tag tree model is used as a base"
        self.assertTrue(issubclass(
            self.tag_model, tag_models.TagTreeModel
        ))

    def test_tree_option(self):
        "Check tag tree model forces tree=True when missing"
        self.assertTrue(self.tag_model.tag_options.tree)
        self.assertTrue(self.singletag_field.tag_options.tree)
        self.assertTrue(self.tag_field.tag_options.tree)

    def test_tree_option_tagmodel_fail(self):
        "Check tree option is not allowed in TagMeta"
        with self.assertRaises(ValueError) as cm:
            class FailModel_tree_tagmeta(tag_models.TagModel):
                class TagMeta:
                    tree = True
        self.assertEqual(
            six.text_type(cm.exception),
            "Cannot set tree option in TagMeta"
        )


###############################################################################
####### TagTreeModel tree navigation methods
###############################################################################

class TagTreeModelNavTest(TagTreeTestManager, TestCase):
    """
    Test navigation through TagTreeModel - get_ancestors etc
    """
    manage_models = [
        test_models.TreeTest,
    ]

    def setUpExtra(self):
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        self.cat = self.tag_model.objects.create(name='Animal/Mammal/Cat')
        self.dog = self.tag_model.objects.create(name='Animal/Mammal/Dog')
        self.bee = self.tag_model.objects.create(name='Animal/Insect/Bee')

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(self.tag_model, {
            'Animal': 0,
            'Animal/Insect': 0,
            'Animal/Insect/Bee': 0,
            'Animal/Mammal': 0,
            'Animal/Mammal/Cat': 0,
            'Animal/Mammal/Dog': 0,
        })

    def test_ancestors(self):
        "Check l1 and l2 ancestors found from l3"
        # Add in some more extras to make sure they're not picked up
        f1 = self.tag_model.objects.create(name='Fail')
        f2 = self.tag_model.objects.create(name='Animal/Fail')
        f3 = self.tag_model.objects.create(name='Animal/Mammal/Cat/Fail')

        # Look up from Cat
        t1 = self.tag_model.objects.get(name='Animal/Mammal/Cat')
        anc = t1.get_ancestors()
        self.assertEqual(len(anc), 2)
        self.assertTreeTag(anc[0], name='Animal', level=1)
        self.assertTreeTag(anc[1], name='Animal/Mammal', level=2)

    def test_no_ancestors(self):
        "Check no ancestors found from l1"
        t1 = self.tag_model.objects.get(name='Animal')
        self.assertEqual(len(t1.get_ancestors()), 0)

    def test_descendants_l1(self):
        "Check l2 and l3 descendants found from l1"
        # Add in some extras to make sure they're not picked up
        f1 = self.tag_model.objects.create(name='Fail')

        # Look down from Animal
        t1 = self.tag_model.objects.get(name='Animal')
        self.assertTreeTag(t1, name='Animal', level=1)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 5)
        self.assertTreeTag(dec[0], name='Animal/Insect', level=2)
        self.assertTreeTag(dec[1], name='Animal/Insect/Bee', level=3)
        self.assertTreeTag(dec[2], name='Animal/Mammal', level=2)
        self.assertTreeTag(dec[3], name='Animal/Mammal/Cat', level=3)
        self.assertTreeTag(dec[4], name='Animal/Mammal/Dog', level=3)

    def test_descendants_l2(self):
        "Check l3 descendants found from l2"
        # Look down from Animal
        t1 = self.tag_model.objects.get(name='Animal/Mammal')
        self.assertTreeTag(t1, name='Animal/Mammal', level=2)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 2)
        self.assertTreeTag(dec[0], name='Animal/Mammal/Cat', level=3)
        self.assertTreeTag(dec[1], name='Animal/Mammal/Dog', level=3)

    def test_descendants_l2(self):
        "Check no descendants found from l3"
        # Look down from Animal
        t1 = self.tag_model.objects.get(name='Animal/Insect/Bee')
        self.assertTreeTag(t1, name='Animal/Insect/Bee', level=3)
        dec = t1.get_descendants()
        self.assertEqual(len(dec), 0)

    def test_siblings_l1(self):
        "Find level 1 siblings"
        # Add another level 1 tag to find
        l1_2 = self.tag_model.objects.create(name='Vegetable')

        t1 = self.tag_model.objects.get(name='Animal')
        sibs = t1.get_siblings()
        self.assertEqual(len(sibs), 2)
        self.assertTreeTag(sibs[0], name='Animal', level=1)
        self.assertTreeTag(sibs[1], name='Vegetable', level=1)

    def test_siblings_l2(self):
        "Find level 2 siblings"
        t1 = self.tag_model.objects.get(name='Animal/Insect')
        sibs = t1.get_siblings()
        self.assertEqual(len(sibs), 2)
        self.assertTreeTag(sibs[0], name='Animal/Insect', level=2)
        self.assertTreeTag(sibs[1], name='Animal/Mammal', level=2)


###############################################################################
####### TagTreeQuerySet
###############################################################################

class TagTreeQuerySetTest(TagTreeTestManager, TestCase):
    """
    Test navigation using TagTreeModelQuerySet
    """
    manage_models = [
        test_models.TreeTest,
    ]

    def setUpExtra(self):
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model

        self.cat = self.tag_model.objects.create(name='Animal/Mammal/Cat')
        self.dog = self.tag_model.objects.create(name='Animal/Mammal/Dog')
        self.bee = self.tag_model.objects.create(name='Animal/Insect/Bee')
        self.vegetable = self.tag_model.objects.create(name='Vegetable')

    def test_setup(self):
        "Check test setup is correct"
        # There should be 6 tags
        self.assertTagModel(self.tag_model, {
            'Animal': 0,
            'Animal/Insect': 0,
            'Animal/Insect/Bee': 0,
            'Animal/Mammal': 0,
            'Animal/Mammal/Cat': 0,
            'Animal/Mammal/Dog': 0,
            'Vegetable': 0,
        })

    def test_ancestors_l1(self):
        qs = self.tag_model.objects.filter(level=1)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Animal', 'Vegetable',
        ])

        qs = qs.with_ancestors()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal', 'Vegetable',
        ])

    def test_ancestors_l2(self):
        qs = self.tag_model.objects.filter(level=2)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Insect', 'Mammal',
        ])

        qs = qs.with_ancestors()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal',
            'Animal/Insect',
            'Animal/Mammal',
        ])

    def test_ancestors_l3(self):
        qs = self.tag_model.objects.filter(level=3)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Bee', 'Cat', 'Dog',
        ])

        qs = qs.with_ancestors()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal',
            'Animal/Insect',
            'Animal/Insect/Bee',
            'Animal/Mammal',
            'Animal/Mammal/Cat',
            'Animal/Mammal/Dog',
        ])

    def test_descendants_l1(self):
        qs = self.tag_model.objects.filter(level=1)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Animal', 'Vegetable',
        ])

        qs = qs.with_descendants()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal',
            'Animal/Insect',
            'Animal/Insect/Bee',
            'Animal/Mammal',
            'Animal/Mammal/Cat',
            'Animal/Mammal/Dog',
            'Vegetable',
        ])

    def test_descendants_l2(self):
        qs = self.tag_model.objects.filter(level=2)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Insect', 'Mammal',
        ])

        qs = qs.with_descendants()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal/Insect',
            'Animal/Insect/Bee',
            'Animal/Mammal',
            'Animal/Mammal/Cat',
            'Animal/Mammal/Dog',
        ])

    def test_descendants_l3(self):
        qs = self.tag_model.objects.filter(level=3)
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Bee', 'Cat', 'Dog',
        ])

        qs = qs.with_descendants()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal/Insect/Bee',
            'Animal/Mammal/Cat',
            'Animal/Mammal/Dog',
        ])

    def test_siblings_l1(self):
        qs = self.tag_model.objects.filter(name='Animal')
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Animal',
        ])

        qs = qs.with_siblings()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal', 'Vegetable',
        ])

    def test_siblings_l2(self):
        # Throw in something else to ensure l2 stays within subtree
        f1 = self.tag_model.objects.create(name='Vegetable/Horrible')

        qs = self.tag_model.objects.filter(name='Animal/Insect')
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Insect',
        ])

        qs = qs.with_siblings()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal/Insect',
            'Animal/Mammal',
        ])

    def test_siblings_l3(self):
        # Throw in something else to ensure l3 stays within subtree
        f1 = self.tag_model.objects.create(name='Vegetable/Horrible/Mushroom')

        qs = self.tag_model.objects.filter(name='Animal/Mammal/Cat')
        self.assertSequenceEqual(qs.values_list('label', flat=True), [
            'Cat',
        ])

        qs = qs.with_siblings()
        self.assertSequenceEqual(qs.values_list('name', flat=True), [
            'Animal/Mammal/Cat',
            'Animal/Mammal/Dog',
        ])



###############################################################################
####### TagTreeModel access via fields
###############################################################################

class TagTreeModelFieldTest(TagTreeTestManager, TestCase):
    """
    Test TagTreeModel access via fields - creation and deletion
    """
    manage_models = [
        test_models.TreeTest,
    ]

    def setUpExtra(self):
        self.singletag_field = test_models.TreeTest.singletag
        self.singletag_model = test_models.TreeTest.singletag.tag_model
        self.tag_field = test_models.TreeTest.tags
        self.tag_model = test_models.TreeTest.tags.tag_model


    def test_singletagfield_level_3(self):
        """
        Check level 3 nodes created with missing level 1 and 2 when assigned
        via singletag fields
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='One/Two/Three',
        )
        t1 = self.singletag_model.objects.get(name='One')
        t2 = self.singletag_model.objects.get(name='One/Two')
        t3 = self.singletag_model.objects.get(name='One/Two/Three')
        self.assertTreeTag(
            t1, name='One', label='One', slug='one', path='one', level=1,
        )
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', label='Three',
            slug='three', path='one/two/three',
            parent=t2, count=1, level=3,
        )

        self.assertTagModel(self.singletag_model, {
            'One': 0,
            'One/Two': 0,
            'One/Two/Three': 1,
        })
        self.assertTagModel(self.tag_model, {})

    def test_tagfield_level_3(self):
        """
        Check level 3 nodes created with missing level 1 and 2 when assigned
        via tag fields
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            tags='One/Two/Three, Uno/Dos/Tres',
        )

        self.assertTagModel(self.singletag_model, {})
        self.assertTagModel(self.tag_model, {
            'One': 0,
            'One/Two': 0,
            'One/Two/Three': 1,
            'Uno': 0,
            'Uno/Dos': 0,
            'Uno/Dos/Tres': 1,
        })

        t1 = self.tag_model.objects.get(name='One')
        t2 = self.tag_model.objects.get(name='One/Two')
        t3 = self.tag_model.objects.get(name='One/Two/Three')
        self.assertTreeTag(
            t1, name='One', label='One', slug='one', path='one', level=1
        )
        self.assertTreeTag(
            t2, name='One/Two', label='Two', slug='two', path='one/two',
            parent=t1, level=2,
        )
        self.assertTreeTag(
            t3, name='One/Two/Three', label='Three',
            slug='three', path='one/two/three',
            parent=t2, level=3, count=1,
        )


    def test_delete_l2_empty_l1(self):
        """
        Check deleting a level 2 node cleans away empty level 1
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='One/Two',
        )
        self.assertTagModel(self.singletag_model, {
            'One': 0,
            'One/Two': 1,
        })

        obj1.singletag = ''
        obj1.save()
        self.assertTagModel(self.singletag_model, {})

    def test_delete_l2_populated_l1(self):
        """
        Check deleting a level 2 node leaves a populated l1
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='Uno/Dos',
        )
        t2 = self.singletag_model.objects.create(name='Uno/Due')
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Dos': 1,
            'Uno/Due': 0,
        })

        obj1.singletag = ''
        obj1.save()
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Due': 0,
        })

    def test_delete_l3_populated_l2(self):
        """
        Check deleting a level 3 node leaves a populated level 2
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='No/Uno/Dos',
        )
        t2 = self.singletag_model.objects.create(name='No/Uno/Due')
        self.assertTagModel(self.singletag_model, {
            'No': 0,
            'No/Uno': 0,
            'No/Uno/Dos': 1,
            'No/Uno/Due': 0,
        })

        obj1.singletag = ''
        obj1.save()
        self.assertTagModel(self.singletag_model, {
            'No': 0,
            'No/Uno': 0,
            'No/Uno/Due': 0,
        })

    def test_delete_l3_empty_l2_populated_l1(self):
        """
        Check deleting a level 3 node cleans empty l2 but leaves populated l1
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='Uno/Dos/Tres',
        )
        t2 = self.singletag_model.objects.create(name='Uno/Due/Tre')
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Dos': 0,
            'Uno/Dos/Tres': 1,
            'Uno/Due': 0,
            'Uno/Due/Tre': 0,
        })

        obj1.singletag = ''
        obj1.save()
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Due': 0,
            'Uno/Due/Tre': 0,
        })

    def test_delete_l3_used_l2(self):
        """
        Check deleting a level 3 node leaves an empty but used level 2
        """
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1',
            singletag='Uno/Dos',
        )
        obj2 = test_models.TreeTest.objects.create(
            name='Test 2',
            singletag='Uno/Dos/Tres',
        )
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Dos': 1,
            'Uno/Dos/Tres': 1,
        })

        obj2.singletag = ''
        obj2.save()
        self.assertTagModel(self.singletag_model, {
            'Uno': 0,
            'Uno/Dos': 1,
        })

    def test_delete_l1_used_l2(self):
        "Check deleting a level 1 node with used level 2 leaves both"
        obj1 = test_models.TreeTest.objects.create(
            name='Test 1', singletag='one',
        )
        obj2 = test_models.TreeTest.objects.create(
            name='Test 2', singletag='one/two',
        )

        self.assertTagModel(self.singletag_model, {
            'one': 1,
            'one/two': 1,
        })
        obj1.singletag = ''
        obj1.save()
        self.assertTagModel(self.singletag_model, {
            'one': 0,
            'one/two': 1,
        })

    def test_get_descendant_count(self):
        "Check the count of the descendants"
        # Make four of everything
        for i in range(4):
            test_models.TreeTest.objects.create(
                name='Test 1.%d' % i, singletag='One',
            )
            test_models.TreeTest.objects.create(
                name='Test 2.%d' % i, singletag='One/Two',
            )
            test_models.TreeTest.objects.create(
                name='Test 3.%d' % i, singletag='One/Two/Three',
            )
        self.assertTagModel(self.singletag_model, {
            'One': 4,
            'One/Two': 4,
            'One/Two/Three': 4,
        })

        # Check the counts
        t1 = self.singletag_model.objects.get(name='One')
        t2 = self.singletag_model.objects.get(name='One/Two')
        t3 = self.singletag_model.objects.get(name='One/Two/Three')
        self.assertEqual(t1.descendant_count, 8)
        self.assertEqual(t2.descendant_count, 4)
        self.assertEqual(t3.descendant_count, 0)

    def test_get_family_count(self):
        "Check the count of the family - node plus descendants"
        # Make four of everything
        for i in range(4):
            test_models.TreeTest.objects.create(
                name='Test 1.%d' % i, singletag='One',
            )
            test_models.TreeTest.objects.create(
                name='Test 2.%d' % i, singletag='One/Two',
            )
            test_models.TreeTest.objects.create(
                name='Test 3.%d' % i, singletag='One/Two/Three',
            )
        self.assertTagModel(self.singletag_model, {
            'One': 4,
            'One/Two': 4,
            'One/Two/Three': 4,
        })

        # Check the counts
        t1 = self.singletag_model.objects.get(name='One')
        t2 = self.singletag_model.objects.get(name='One/Two')
        t3 = self.singletag_model.objects.get(name='One/Two/Three')
        self.assertEqual(t1.family_count, 12)
        self.assertEqual(t2.family_count, 8)
        self.assertEqual(t3.family_count, 4)



"""
Tagulous test: Model TagField

Modules tested:
    tagulous.models.managers.TagRelatedManager
    tagulous.models.descriptors.BaseTagDescriptor
    tagulous.models.descriptors.TagDescriptor
    tagulous.models.fields.BaseTagField
    tagulous.models.fields.TagField
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

try:
    from django.core.checks import Warning as ChecksWarning
except ImportError:
    ChecksWarning = None

from tests.lib import *


class ModelTagFieldTest(TagTestManager, TestCase):
    """
    Test model TagField
    """
    manage_models = [
        test_models.TagFieldModel,
    ]

    def setUpExtra(self):
        self.tag_model = test_models.TagFieldModel.tags.tag_model
        self.tag_field = test_models.TagFieldModel.tags

    def test_descriptor(self):
        "Check TagDescriptor is in place"
        self.assertIsInstance(
            self.tag_field, tag_models.TagDescriptor
        )

    def test_tag_table(self):
        "Check the tag table exists"
        if django.VERSION < (1, 9):
            self.assertTrue(issubclass(
                self.tag_field.field.remote_field.to, tag_models.TagModel,
            ))
        else:
            self.assertTrue(issubclass(
                self.tag_field.field.remote_field.model, tag_models.TagModel,
            ))
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))

    def test_empty_value(self):
        "Check the descriptor returns a TagRelatedManager"
        t1 = self.create(test_models.TagFieldModel, name="Test")
        self.assertInstanceEqual(t1, name="Test")

        # Check the TagDescriptor is returning a TagRelatedManager
        self.assertEqual(t1.tags.__class__.__name__, 'TagRelatedManager')

        # Check it also has a reference to the correct model
        self.assertEqual(t1.tags.tag_model, self.tag_model)

    def test_tag_assign_before_save(self):
        """
        Check a tag string can be assigned to an instance which hasn't yet
        been saved
        """
        t1 = test_models.TagFieldModel(name="Test")
        t1.tags = 'blue, red'
        self.assertEqual(t1.tags.get_tag_string(), 'blue, red')
        self.assertTagModel(self.tag_model, {})

    def test_tag_assign_in_constructor(self):
        "Check a tag string can be set in the constructor"
        t1 = test_models.TagFieldModel(name="Test", tags='blue, red')

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), 'blue, red')
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertInstanceEqual(t1, name='Test', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue': 1,
            'red':  1,
        })

    def test_tag_assign_by_list_in_constructor(self):
        "Check a list of strings can be set in the constructor"
        t1 = test_models.TagFieldModel(name="Test", tags=['blue', 'red'])

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), 'blue, red')
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertInstanceEqual(t1, name='Test', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_in_object_create(self):
        "Check a tag string can be set in object.create"
        t1 = test_models.TagFieldModel.objects.create(
            name="Test", tags='blue, red',
        )
        self.assertInstanceEqual(t1, name='Test', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_by_list_in_object_create(self):
        "Check a list of strings can be set in object.create"
        t1 = test_models.TagFieldModel.objects.create(
            name="Test", tags=['blue', 'red']
        )
        self.assertInstanceEqual(t1, name='Test', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_by_queryset_in_object_create(self):
        "Check a list of strings can be set in object.create"
        self.tag_model.objects.create(name='blue')
        self.tag_model.objects.create(name='red')
        t1 = test_models.TagFieldModel.objects.create(
            name="Test", tags=self.tag_model.objects.all()
        )
        self.assertInstanceEqual(t1, name='Test', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_in_object_get_or_create_true(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does not exist
        """
        t1, state = test_models.TagFieldModel.objects.get_or_create(
            name='Test', tags='blue, red',
        )
        self.assertEqual(state, True)
        self.assertEqual(t1.name, 'Test')
        self.assertEqual(t1.tags, 'blue, red')
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_in_object_get_or_create_false(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does exist
        """
        t1 = test_models.TagFieldModel.objects.create(
            name="Test", tags='blue, red',
        )
        t2 = test_models.TagFieldModel.objects.create(
            name="Test", tags='green',
        )
        t3, state = test_models.TagFieldModel.objects.get_or_create(
            name='Test', tags='blue, red',
        )
        self.assertEqual(state, False)
        self.assertEqual(t1.pk, t3.pk)
        self.assertInstanceEqual(t2, name="Test", tags='green')
        self.assertEqual(t3.name, 'Test')
        self.assertEqual(t3.tags, 'blue, red')
        self.assertTagModel(self.tag_model, {
            'red':   1,
            'blue':  1,
            'green': 1,
        })

    def test_tag_assign_string_obj_save(self):
        """
        Check a tag string can be assigned to the descriptor, saved when the
        instance is saved, and returned

        Checks TagRelatedManager post_save listener
        """
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        t1.tags = 'red, blue'

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), 'blue, red')
        self.assertEqual(six.text_type(t1.tags), t1.tags.get_tag_string())
        self.assertEqual(len(t1.tags.get_tag_list()), 2)
        self.assertTrue('red' in t1.tags.get_tag_list())
        self.assertTrue('blue' in t1.tags.get_tag_list())
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), '')

        # Returned after save
        t1.save()
        t1 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(t1.tags.get_tag_string(), 'blue, red')
        self.assertEqual(six.text_type(t1.tags), t1.tags.get_tag_string())
        self.assertEqual(len(t1.tags.get_tag_list()), 2)
        self.assertTrue('red' in t1.tags.get_tag_list())
        self.assertTrue('blue' in t1.tags.get_tag_list())
        self.assertTagModel(self.tag_model, {
            'red':  1,
            'blue': 1,
        })

    def test_tag_assign_string_manager_save(self):
        """
        Check a tag string can be assigned to the descriptor, saved by the tag
        manager, and returned

        Checks TagRelatedManager.save
        """
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        t1.tags = 'red, blue'

        # Returned before save
        self.assertEqual(six.text_type(t1.tags), 'blue, red')
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), '')

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

    def test_assign_list_strings(self):
        "Check tags can be set with a list of strings"
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        t1.tags = ['red', 'blue']

        # Returned before save
        self.assertEqual(six.text_type(t1.tags), 'blue, red')
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), '')

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

    def test_assign_list_objs(self):
        "Check tags can be set with a list of strings"
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        t1.tags = [
            self.tag_model.objects.create(name='blue'),
            self.tag_model.objects.create(name='red')
        ]

        # Returned before save
        self.assertEqual(six.text_type(t1.tags), 'blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      0,
            'blue':     0,
        })

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), '')

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

    def test_assign_queryset(self):
        "Check tags can be set with a queryset of tags"
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        self.tag_model.objects.create(name='blue')
        self.tag_model.objects.create(name='red')
        t1.tags = self.tag_model.objects.all()

        # Returned before save, check DB hasn't changed
        self.assertEqual(six.text_type(t1.tags), 'blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertTagModel(self.tag_model, {
            'red':      0,
            'blue':     0,
        })

        # Returned after save
        t1.tags.save()
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

    def test_tag_assign_string_empty(self):
        "Check setting an empty string clears tags"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })
        t1.tags = ''

        # Returned before save, check DB hasn't changed
        self.assertEqual(six.text_type(t1.tags), '')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertTagModel(self.tag_model, {})

    def test_assign_list_empty(self):
        "Check setting an empty list clears tags"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })
        t1.tags = []

        # Returned before save
        self.assertEqual(six.text_type(t1.tags), '')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), 'blue, red')

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertTagModel(self.tag_model, {})

    def test_assign_none(self):
        "Check setting None clears tags"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })
        t1.tags = None

        # Returned before save
        self.assertEqual(six.text_type(t1.tags), '')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

        # Check db hasn't changed
        t2 = test_models.TagFieldModel.objects.get(name='Test 1')
        self.assertEqual(six.text_type(t2.tags), 'blue, red')

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertTagModel(self.tag_model, {})

    def test_assign_unknown(self):
        "Check assigning an unknown value raises an exception"
        t1 = self.create(test_models.TagFieldModel, name="Test 1")
        with self.assertRaises(ValueError) as cm:
            # Assign the descriptor to itself, because why not.
            t1.tags = test_models.TagFieldModel.tags
        self.assertEqual(
            six.text_type(cm.exception), 'Unexpected value assigned to TagField'
        )

    def test_contains(self):
        "Check __contains__ using strings and Tag objects"
        t1 = self.create(test_models.TagFieldModel, name="1", tags='blue, red')

        # String matches
        self.assertTrue('blue' in t1.tags, 'blue not found')
        self.assertTrue('red' in t1.tags, 'red not found')

        # Partials shouldn't match
        self.assertFalse('green' in t1.tags, 'green incorrectly found')
        self.assertFalse('blu' in t1.tags, 'blu incorrectly found')
        self.assertFalse('ed' in t1.tags, 'ed incorrectly found')

        # Tag objects (will be cast to strings)
        red = t1.tags.tag_model.objects.get(name='red')
        blue = t1.tags.tag_model.objects.get(name='blue')
        green = t1.tags.tag_model.objects.create(name='green')
        self.assertTrue(red in t1.tags, 'red not found')
        self.assertTrue(blue in t1.tags, 'red not found')
        self.assertFalse(green in t1.tags, 'green incorrectly found')

    def test_len(self):
        "Check manager count method returns the number of tags"
        t1 = self.create(test_models.TagFieldModel, name="1", tags='blue, red')
        self.assertEqual(t1.tags.count(), 2)

    def test_eq(self):
        "Check __eq__ correctly determines equality"
        t1 = self.create(test_models.TagFieldModel, name="1", tags='blue, red')
        self.assertEqual(t1.tags, 'blue, red')
        self.assertEqual(t1.tags, 'red, blue')
        self.assertEqual(t1.tags, ['red', 'blue'])

        # Test queryset
        tags = t1.tags.tag_model.objects.all()
        self.assertEqual(t1.tags, tags)

        # Test another TagRelatedManager
        t2 = self.create(test_models.TagFieldModel, name="2", tags='blue, red')
        self.assertEqual(t1.tags, t2.tags)

    def test_ne(self):
        "Check __ne__ correctly determines falsity"
        t1 = self.create(test_models.TagFieldModel, name="1", tags='blue, red')
        self.assertNotEqual(t1.tags, 'red')
        self.assertNotEqual(t1.tags, 'blue')
        self.assertNotEqual(t1.tags, ['red', 'green'])
        self.assertNotEqual(t1.tags, 'green')

        # Test another TagRelatedManager
        t2 = self.create(test_models.TagFieldModel, name="2", tags='red, green')
        self.assertNotEqual(t1.tags, t2.tags)

        # Test queryset
        tags = t1.tags.tag_model.objects.all()
        self.assertNotEqual(t1.tags, tags)


    def test_multiple_instances(self):
        "Check multiple tagged instances work without interference"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='green, red')
        t3 = self.create(test_models.TagFieldModel, name="Test 3", tags='blue, green')

        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertInstanceEqual(t2, name='Test 2', tags='green, red')
        self.assertInstanceEqual(t3, name='Test 3', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'red':      2,
            'blue':     2,
            'green':    2,
        })

    def test_change_string_add(self):
        "Add a tag by changing tag string"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
        })
        t1.tags = 'blue, green'
        self.assertTagModel(self.tag_model, {
            'blue':     1,
        })
        t1.tags.save()
        self.assertEqual(t1.tags, 'blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

    def test_m2m_set_by_string(self):
        "Set a tag directly using M2M .set(string)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'red':      1,
        })
        t1.tags.set('green', 'pink')
        self.assertEqual(t1.tags, 'green, pink')
        self.assertInstanceEqual(t1, name='Test 1', tags='green, pink')
        self.assertTagModel(self.tag_model, {
            'green':    1,
            'pink':     1,
        })

        # No need to test further arguments, it passes through to add() which
        # is tested thoroughly below

    def test_m2m_add_by_obj(self):
        "Add a tag directly using M2M .add(obj)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
        })
        t1.tags.add(
            self.tag_model.objects.create(name='green')
        )
        self.assertEqual(t1.tags, 'blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

    def test_m2m_add_by_string(self):
        "Add a tag directly using M2M .add(str)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
        })
        t1.tags.add('green')
        self.assertEqual(t1.tags, 'blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

    def test_m2m_add_by_string_existing(self):
        "Add a tag directly using M2M .add(str) when the tag already exists"
        test_models.TagFieldModel.tags.tag_model.objects.create(name='green')
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    0,
        })
        t1.tags.add('green')
        self.assertEqual(t1.tags, 'blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

    def test_change_string_remove(self):
        "Remove a tag by changing tag string"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })

        t1.tags = 'green'
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })
        t1.tags.save()
        self.assertInstanceEqual(t1, name='Test 1', tags='green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
            'red':      1,
        })

    def test_m2m_remove_by_obj(self):
        "Remove a tag directly using M2M .remove(obj)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })

        t1.tags.remove(
            self.tag_model.objects.get(name='blue')
        )
        self.assertEqual(t1.tags, 'green')
        self.assertInstanceEqual(t1, name='Test 1', tags='green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
            'red':      1,
        })

    def test_m2m_remove_by_str(self):
        "Remove a tag directly using M2M .remove(str)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })

        t1.tags.remove('blue')
        self.assertEqual(t1.tags, 'green')
        self.assertInstanceEqual(t1, name='Test 1', tags='green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
            'red':      1,
        })

    def test_m2m_remove_not_set(self):
        "Remove a tag which exists but isn't set"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })

        t1.tags.remove('blue', 'green', 'red')
        self.assertEqual(t1.tags, '')
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'red':      1,
        })

    def test_m2m_remove_missing(self):
        "Remove a tag which doesn't exist (and isn't set)"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        t1.tags.remove('green')
        self.assertEqual(t1.tags, 'blue, red')

    def test_m2m_clear(self):
        "Clear all tags from an item with manager.clear"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      2,
            'blue':     2,
        })

        # Returned before save
        t1.tags.clear()
        self.assertEqual(six.text_type(t1.tags), '')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')

    def test_fake_manager_after_delete(self):
        "Check fake manager steps in to store tags when the item is deleted"
        # Create the item
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'red':      1,
            'blue':     1,
        })

        # Now delete it
        t1.delete()

        # It won't have a pk any more, but it will have a manager.
        self.assertEqual(t1.pk, None)

        # TagDescriptor will now have to create a FakeTagRelatedManager and
        # copy the tags across from the real one. Because although this is
        # about as likely as the previous test, someone will do it eventually.
        manager = t1.tags
        self.assertIsInstance(
            manager, tag_models.managers.FakeTagRelatedManager
        )
        self.assertEqual(manager, 'blue, red')

    def test_fake_manager(self):
        "Check that the FakeTagRelatedManager doesn't do databases"
        if django.VERSION < (2, 0):
            errmsg = (
                '"<TagFieldModel: TagFieldModel object>" needs to be '
                'saved before TagField can use the database'
            )
        else:
            # Django 2.0 adds primary key to model __str__ and __repr__
            errmsg = (
                '"<TagFieldModel: TagFieldModel object (None)>" needs to be '
                'saved before TagField can use the database'
            )

        t1 = test_models.TagFieldModel(name="Test 1", tags='blue, red')

        with self.assertRaises(ValueError) as cm:
            t1.tags.save()
        self.assertEqual(six.text_type(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.set('fail')
        self.assertEqual(six.text_type(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.add('fail')
        self.assertEqual(six.text_type(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.remove('blue')
        self.assertEqual(six.text_type(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.clear()
        self.assertEqual(six.text_type(cm.exception), errmsg)


###############################################################################
#######  Test invalid fields
###############################################################################

class ModelTagFieldInvalidTest(TagTestManager, TransactionTestCase):
    """
    Test invalid model TagField

    Use a TransactionTestCase so the apps will be reset
    """
    manage_models = [
        test_models.SingleTagFieldModel,
    ]

    def test_to_model_options_fails(self):
        "Check when a to model is specified, tag options are invalid"
        new_field = tag_models.TagField(
            test_models.TagMetaUser, force_lowercase=True,
        )
        with self.assertRaises(ValueError) as cm:
            new_field._process_deferred_options()
        self.assertEqual(
            six.text_type(cm.exception),
            "Cannot set tag options on explicit tag model "
            "<class 'tests.tagulous_tests_app.models.TagMetaUser'>"
        )

    def test_to_model_force_options_works(self):
        "Check when _set_tag_meta is True, options can be set after all"
        self.assertTrue(test_models.TagMetaModel.tag_options.force_lowercase)
        new_field = tag_models.TagField(
            test_models.TagMetaModel, force_lowercase=False,
            _set_tag_meta=True,
        )
        new_field._process_deferred_options()
        self.assertFalse(test_models.TagMetaModel.tag_options.force_lowercase)

        # Set force_lowercase back ready for future tests - unittest won't fix
        test_models.TagMetaModel.tag_options.force_lowercase = True
        self.assertTrue(test_models.TagMetaModel.tag_options.force_lowercase)

    def test_contribute_only_once(self):
        "Check that a field can't be contributed more than once"
        field = test_models.TagFieldModel._meta.get_field('tags')
        with self.assertRaises(AttributeError) as cm:
            field.contribute_to_class(test_models.TagFieldModel, 'new_tags')
        self.assertEqual(
            six.text_type(cm.exception),
            "The tag field <tagulous.models.fields.TagField: tags> is already "
            "attached to a model"
        )

    def test_invalid_to_model(self):
        "Check that the to model has to be a TagModel subclass"
        with self.assertRaises(ValueError) as cm:
            class FailModel_invalid_to(models.Model):
                to_model = tag_models.TagField(test_models.TagFieldModel)
        self.assertEqual(
            six.text_type(cm.exception),
            "Tag model must be a subclass of TagModel"
        )

    def test_forbidden_db_table(self):
        "Check that db_table argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_db(models.Model):
                db_table = tag_models.TagField(db_table='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'db_table' for TagField"
        )

    def test_forbidden_through(self):
        "Check that through argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_through(models.Model):
                through = tag_models.TagField(through='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'through' for TagField"
        )

    def test_forbidden_symmetrical(self):
        "Check that symmetrical argument raises exception"
        with self.assertRaises(ValueError) as cm:
            class FailModel_forbidden_symmetrical(models.Model):
                symmetrical = tag_models.TagField(symmetrical='fail')
        self.assertEqual(
            six.text_type(cm.exception),
            "Invalid argument 'symmetrical' for TagField"
        )

    @unittest.skipIf(ChecksWarning is None, 'Check test only run for Django 1.9+')
    def test_nulled_tag_field(self):
        "Check model field raises warning when given invalid null"
        nulled_tagfield = tag_models.TagField(null=True)
        warnings = nulled_tagfield._check_ignored_options()
        self.assertEqual(len(warnings), 1)
        self.assertEqual(warnings[0].msg, 'null has no effect on TagField.')


###############################################################################
####### Test tag counter
###############################################################################

class ModelTagFieldCountTest(TagTestManager, TestCase):
    """
    Test tag model counter
    """
    manage_models = [
        test_models.TagFieldModel,
    ]

    def setUpExtra(self):
        self.tag_model = test_models.TagFieldModel.tags.tag_model
        self.tag_field = test_models.TagFieldModel.tags

    def test_count_0_create(self):
        "Check an unprotected tag can be created with a tag count of 0"
        self.tag_model.objects.create(name='red')
        self.assertTagModel(self.tag_model, {
            'red':      0,
        })

    def test_count_0_delete(self):
        "Check that an unprotected tag is deleted when count changes to 0"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

        t1.tags.remove('blue')
        self.assertInstanceEqual(t1, name='Test 1', tags='green')
        self.assertTagModel(self.tag_model, {
            'green':    1,
        })

    def test_instance_delete_updates(self):
        "Check that deleting an instance updates the count"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags='blue, red')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     2,
            'green':    1,
            'red':      1,
        })

        t1.delete()
        self.assertInstanceEqual(t2, name='Test 2', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'red':      1,
        })

    def test_protected_remains(self):
        "Check that a protected tag stays when count changes to 0"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

        green = self.tag_model.objects.get(name='green')
        green.protected = True
        green.save()

        t1.delete()
        self.assertTagModel(self.tag_model, {
            'green':    0,
        })

    def test_clear_count_0_delete(self):
        """
        Check that an unprotected tag is deleted when clear changes count to 0,
        but protected tag remains
        """
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags='blue, green')
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, green')
        self.assertTagModel(self.tag_model, {
            'blue':     1,
            'green':    1,
        })

        green = self.tag_model.objects.get(name='green')
        green.protected = True
        green.save()

        t1.tags.clear()
        self.assertInstanceEqual(t1, name='Test 1', tags='')
        self.assertTagModel(self.tag_model, {
            'green':    0,
        })


###############################################################################
#######  Test model field blank=True
###############################################################################

class ModelTagFieldOptionalTest(TagTestManager, TestCase):
    """
    Test optional TagField
    """
    manage_models = [
        test_models.TagFieldOptionalModel,
    ]
    def test_optional_save_missing(self):
        "Check an optional TagField isn't required for save"
        # If it fails, it will return an exception
        t1 = test_models.TagFieldOptionalModel(name='Test 1')
        t1.save()
        self.assertNotEqual(t1.pk, None)

    def test_optional_create_missing(self):
        "Check an optional TagField isn't required for object.create"
        # If it fails, it will return an exception
        t1 = test_models.TagFieldOptionalModel.objects.create(name='Test 1')
        self.assertNotEqual(t1.pk, None)


###############################################################################
####### Test model field blank=False
###############################################################################

class ModelTagFieldRequiredTest(TagTestManager, TestCase):
    """
    Test required model TagField
    """
    manage_models = [
        test_models.TagFieldRequiredModel,
    ]
    def test_required_save(self):
        "Check a required TagField doesn't raise an exception when saved"
        # This isn't exactly the desired action, but as it stands we can't
        # enforce blank=False on a M2M tagfield without rewriting chunks of
        # django's core modelform - it can't even be monkeypatched cleanly.
        #
        # Future changes to django may make it possible; at that point we can
        # switch this test to enforce blank=False:
        '''
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.TagFieldRequiredModel(name='Test')
            t1.save()
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')
        '''
        # but until then:
        t1 = test_models.TagFieldRequiredModel(name='Test')
        t1.save()
        self.assertInstanceEqual(t1, name='Test', tag='')

    def test_required_create_raises(self):
        "Check a required TagField raises an exception in object.create"
        # As test_required_save, this isn't desired behaviour, but until django
        # changes, we can't use this test:
        '''
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.TagFieldRequiredModel.objects.create(name='Test')
            t1.full_clean()
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')
        '''
        # instead:
        t1 = test_models.TagFieldRequiredModel.objects.create(name='Test')
        self.assertInstanceEqual(t1, name='Test', tag='')


###############################################################################
####### Test multiple TagFields on a model
###############################################################################

class ModelTagFieldMultipleTest(TagTestManager, TestCase):
    """
    Test multiple tag fields on a model
    """
    manage_models = [
        test_models.TagFieldMultipleModel,
    ]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldMultipleModel
        self.tag_field_1 = test_models.TagFieldMultipleModel.tags1
        self.tag_field_2 = test_models.TagFieldMultipleModel.tags2
        self.tag_field_3 = test_models.TagFieldMultipleModel.tags3

    def test_tag_models(self):
        self.assertTrue(issubclass(self.tag_field_1.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_2.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_3.tag_model, tag_models.TagModel))

    def test_separate_models(self):
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_2.tag_model)
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_3.tag_model)
        self.assertNotEqual(self.tag_field_2.tag_model, self.tag_field_3.tag_model)

    def test_model_names(self):
        self.assertEqual(
            self.tag_field_1.tag_model.__name__,
            'Tagulous_TagFieldMultipleModel_tags1'
        )
        self.assertEqual(
            self.tag_field_2.tag_model.__name__,
            'Tagulous_TagFieldMultipleModel_tags2'
        )
        self.assertEqual(
            self.tag_field_3.tag_model.__name__,
            'Tagulous_TagFieldMultipleModel_tags3'
        )

    def test_set_and_get(self):
        "Test multiple fields can be set and retrieved independently"
        t1 = self.create(
            self.test_model, name="Test 1",
            tags1='adam, brian',
            tags2='alice, brenda',
            tags3='aragorn, bilbo',
        )
        t2 = self.create(
            self.test_model, name="Test 2",
            tags1='chris, david',
            tags2='claire, daphne',
            tags3='celeborn, denethor',
        )
        t3 = self.create(
            self.test_model, name="Test 3",
            tags1='eric, frank',
            tags2='edith, faye',
            tags3='eowyn, frodo',
        )

        self.assertTagModel(self.tag_field_1, {
            'adam':     1,  'brian':    1,  'chris':    1,
            'david':    1,  'eric':     1,  'frank':    1,
        })
        self.assertTagModel(self.tag_field_2, {
            'alice':    1,  'brenda':   1,  'claire':   1,
            'daphne':   1,  'edith':    1,  'faye':     1,
        })
        self.assertTagModel(self.tag_field_3, {
            'aragorn':  1,  'bilbo':    1,  'celeborn': 1,
            'denethor': 1,  'eowyn':    1,  'frodo':    1,
        })

        self.assertInstanceEqual(
            t1, name="Test 1",
            tags1='adam, brian',
            tags2='alice, brenda',
            tags3='aragorn, bilbo',
        )
        self.assertInstanceEqual(
            t2, name="Test 2",
            tags1='chris, david',
            tags2='claire, daphne',
            tags3='celeborn, denethor',
        )
        self.assertInstanceEqual(
            t3, name="Test 3",
            tags1='eric, frank',
            tags2='edith, faye',
            tags3='eowyn, frodo',
        )



###############################################################################
####### Test TagField with string references to tag model
###############################################################################

class ModelTagFieldStringTest(TagTestManager, TransactionTestCase):
    """
    Test tag field which refers to its tag model with a string
    """
    manage_models = [
        test_models.MixedStringTo,
    ]

    def setUpExtra(self):
        self.test_model = test_models.MixedStringTo
        self.tag_field = self.test_model.tags
        self.tag_model = test_models.MixedStringTagModel

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        if django.VERSION < (1, 9):
            self.assertEqual(self.tag_field.field.remote_field.to, self.tag_model)
        else:
            self.assertEqual(self.tag_field.field.remote_field.model, self.tag_model)
        self.assertEqual(self.tag_field.tag_model, self.tag_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(
            self.tag_field.tag_options, self.tag_model.tag_options
        )

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.tag_model, {})
        t1 = self.test_model.objects.create(name='Test 1', tags='one, two')
        self.assertTagModel(self.tag_model, {
            'one':  1,
            'two':  1,
        })
        self.assertInstanceEqual(t1, name='Test 1', tags='one, two')


class ModelTagFieldSelfTest(TagTestManager, TransactionTestCase):
    """
    Test tag field which refers to itself
    """
    manage_models = [
        test_models.MixedSelfTo,
    ]

    def setUpExtra(self):
        self.test_model = test_models.MixedSelfTo
        self.tag_field = self.test_model.related

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        if django.VERSION < (1, 9):
            self.assertEqual(self.tag_field.field.remote_field.to, self.test_model)
        else:
            self.assertEqual(self.tag_field.field.remote_field.model, self.test_model)
        self.assertEqual(self.tag_field.tag_model, self.test_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(
            self.tag_field.tag_options, self.test_model.tag_options
        )

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.test_model, {})
        t1 = self.test_model.objects.create(name='Test 1', related='one, two')
        self.assertTagModel(self.test_model, {
            'one':  1,
            'two':  1,
            'Test 1': 0,
        })
        self.assertInstanceEqual(t1, name='Test 1', related='one, two')



###############################################################################
####### Test TagField options
###############################################################################

class ModelTagFieldOptionsTest(TagTestManager, TransactionTestCase):
    """
    Test tag field options
    """
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldOptionsModel

    def test_initial_string(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(self.test_model.initial_string, {
            'Adam':     0,
            'Brian':    0,
            'Chris':    0,
        })

    def test_initial_list(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(self.test_model.initial_list, {
            'Adam':     0,
            'Brian':    0,
            'Chris':    0,
        })

    def test_protect_initial_true(self):
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Adam':     0,
        })

        t1 = self.create(self.test_model, name="Test 1", protect_initial_true='Adam')
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Adam':     1,
        })

        t1.protect_initial_true = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Adam':     0,
        })

    def test_protect_initial_false(self):
        self.assertTagModel(self.test_model.protect_initial_true, {
            'Adam':     0,
        })

        t1 = self.create(self.test_model, name="Test 1", protect_initial_false='Adam')
        self.assertTagModel(self.test_model.protect_initial_false, {
            'Adam':     1,
        })

        t1.protect_initial_false = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_false, {})

    def test_protect_all_true(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_true='Adam')
        self.assertTagModel(self.test_model.protect_all_true, {
            'Adam':     1,
        })

        t1.protect_all_true = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_all_true, {
            'Adam':     0,
        })

    def test_protect_all_false(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_false='Adam')
        self.assertTagModel(self.test_model.protect_all_false, {
            'Adam':     1,
        })

        t1.protect_all_false = ''
        t1.save()
        self.assertTagModel(self.test_model.protect_all_false, {})

    @skip_if_mysql
    def test_case_sensitive_true(self):
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Adam':     0,
        })
        t1 = self.create(self.test_model, name="Test 1", case_sensitive_true='adam')
        self.assertTagModel(self.test_model.case_sensitive_true, {
            'Adam':     0,
            'adam':     1,
        })

    @skip_if_mysql
    def test_case_sensitive_false(self):
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Adam':     0,
        })
        t1 = self.create(self.test_model, name="Test 1", case_sensitive_false='adam')
        self.assertTagModel(self.test_model.case_sensitive_false, {
            'Adam':     1,
        })

    @skip_if_mysql
    def test_cmp_case_sensitive_true(self):
        "Test case sensitive matches"
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_true='django, html',
        )
        self.assertEqual(t1.case_sensitive_true, 'django, html')
        self.assertNotEqual(t1.case_sensitive_true, 'django, HTML')
        self.assertNotEqual(t1.case_sensitive_true, 'Django, html')

    @skip_if_mysql
    def test_contains_case_sensitive_true(self):
        "Check case sensitive __contains__"
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_true='Django, HTML',
        )

        # String matches
        self.assertTrue('Django' in t1.case_sensitive_true)
        self.assertTrue('HTML' in t1.case_sensitive_true)

        # Partials and incorrect shouldn't be found
        self.assertFalse('Javascript' in t1.case_sensitive_true)
        self.assertFalse('django' in t1.case_sensitive_true)
        self.assertFalse('Html' in t1.case_sensitive_true)

    @skip_if_mysql
    def test_cmp_case_sensitive_false(self):
        """
        Test case insensitive matches

        Note: force_lowercase is also false here
        """
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_false='Adam, Brian',
        )
        self.assertEqual(t1.case_sensitive_false, 'Adam, Brian')
        self.assertEqual(t1.case_sensitive_false, 'adam, BRIAN')
        self.assertNotEqual(t1.case_sensitive_false, 'Chris')

    def test_force_lowercase_true(self):
        t1 = self.create(self.test_model, name="Test 1", force_lowercase_true='Adam')
        self.assertTagModel(self.test_model.force_lowercase_true, {
            'adam':     1,
        })

    def test_force_lowercase_false(self):
        t1 = self.create(self.test_model, name="Test 1", force_lowercase_false='Adam')
        self.assertTagModel(self.test_model.force_lowercase_false, {
            'Adam':     1,
        })

    def test_cmp_force_lowercase_true(self):
        "Test matches when force lowercase true and case sensitive is false"
        t1 = self.create(
            self.test_model, name="Test 1", force_lowercase_true='blue, red',
        )
        self.assertEqual(t1.force_lowercase_true, 'blue, red')
        self.assertEqual(t1.force_lowercase_true, 'Blue, RED')
        self.assertNotEqual(t1.force_lowercase_true, 'green')

    def test_contains_force_lowercase_true(self):
        "Check __contains__ when force lowercase true and case sensitive false"
        t1 = self.create(
            self.test_model, name="Test 1", force_lowercase_true='django, html',
        )

        # String matches are forced to lowercase
        self.assertTrue('Django' in t1.force_lowercase_true)
        self.assertTrue('HTML' in t1.force_lowercase_true)
        self.assertTrue('django' in t1.force_lowercase_true)
        self.assertTrue('html' in t1.force_lowercase_true)

        # Incorrect shouldn't be found
        self.assertFalse('Javascript' in t1.case_sensitive_true)

    def test_cmp_case_sensitive_true_force_lowercase_true(self):
        "Test matches when case sensitive and force lowercase true"
        t1 = self.create(
            self.test_model, name="Test 1",
            cs_true_fl_true='blue, red',
        )
        self.assertEqual(t1.cs_true_fl_true, 'blue, red')
        self.assertEqual(t1.cs_true_fl_true, 'Blue, RED')
        self.assertNotEqual(t1.cs_true_fl_true, 'green')

    def test_max_count_below(self):
        t1 = self.create(self.test_model, name="Test 1", max_count='Adam')
        self.assertInstanceEqual(t1, name="Test 1", max_count='Adam')
        self.assertTagModel(self.test_model.max_count, {
            'Adam':     1,
        })

    def test_max_count_equal(self):
        t1 = self.create(self.test_model, name="Test 1", max_count='Adam, Brian, Chris')
        self.assertInstanceEqual(t1, name="Test 1", max_count='Adam, Brian, Chris')
        self.assertTagModel(self.test_model.max_count, {
            'Adam':     1,
            'Brian':    1,
            'Chris':    1,
        })

    def test_max_count_create_above(self):
        with self.assertRaises(ValueError) as cm:
            t1 = self.test_model.objects.create(
                name="Test 1",
                max_count='Adam, Brian, Chris, David',
            )
        self.assertEqual(
            six.text_type(cm.exception),
            "Cannot set more than 3 tags on this field"
        )
        with self.assertRaises(self.test_model.DoesNotExist) as cm:
            t2 = self.test_model.objects.get(name="Test 1")
        self.assertTagModel(self.test_model.max_count, {})

    def test_max_count_assign_equal_replaces(self):
        t1 = self.create(self.test_model, name="Test 1", max_count='Adam, Brian, Chris')
        t1.max_count = 'Adam, Brian, David'
        t1.save()
        self.assertInstanceEqual(t1, name="Test 1", max_count='Adam, Brian, David')
        self.assertTagModel(self.test_model.max_count, {
            'Adam':     1,
            'Brian':    1,
            'David':    1,
        })

    def test_max_count_assign_above(self):
        t1 = self.create(self.test_model, name="Test 1")
        with self.assertRaises(ValueError) as cm:
            t1.max_count = 'Adam, Brian, Chris, David'
        self.assertEqual(
            six.text_type(cm.exception),
            "Cannot set more than 3 tags on this field"
        )
        self.assertInstanceEqual(t1, name="Test 1", max_count='')
        self.assertTagModel(self.test_model.max_count, {})

    def test_max_count_add_above(self):
        t1 = self.create(self.test_model, name="Test 1", max_count='Adam, Brian')
        with self.assertRaises(ValueError) as cm:
            t1.max_count.add('Chris', 'David')
        self.assertEqual(
            six.text_type(cm.exception),
            "Cannot set more than 3 tags on this field; it already has 2"
        )
        self.assertInstanceEqual(t1, name="Test 1", max_count='Adam, Brian')
        # They won't have been created
        self.assertTagModel(self.test_model.max_count, {
            'Adam':     1,
            'Brian':    1,
        })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils import six

import tagulous.models.fields
import tagulous.models.models


class Migration(migrations.Migration):

    dependencies = [
        ('tagulous_tests_migration', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tagulous_MigrationTestModel_singletag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=191)),
                ('slug', models.SlugField()),
                ('count', models.IntegerField(default=0, help_text=b'Internal counter of how many times this tag is in use')),
                ('protected', models.BooleanField(default=False, help_text=b'Will not be deleted when the count reaches 0')),
            ],
            options={
                'ordering': ('name',),
                'abstract': False,
            },
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.CreateModel(
            name='Tagulous_MigrationTestModel_tags',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=191)),
                ('slug', models.SlugField()),
                ('count', models.IntegerField(default=0, help_text=b'Internal counter of how many times this tag is in use')),
                ('protected', models.BooleanField(default=False, help_text=b'Will not be deleted when the count reaches 0')),
            ],
            options={
                'ordering': ('name',),
                'abstract': False,
            },
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name='tagulous_migrationtestmodel_tags',
            unique_together=set([('slug',)]),
        ),
        migrations.AlterUniqueTogether(
            name='tagulous_migrationtestmodel_singletag',
            unique_together=set([('slug',)]),
        ),
        migrations.AddField(
            model_name='migrationtestmodel',
            name='singletag',
            field=tagulous.models.fields.SingleTagField(force_lowercase=False, to='tagulous_tests_migration.Tagulous_MigrationTestModel_singletag', blank=True, help_text=b'Enter a comma-separated tag string', null=True, autocomplete_initial=True, autocomplete_view=b'tagulous_tests_app-null', autocomplete_limit=3, initial='Mr', protect_all=False, case_sensitive=True, protect_initial=True, _set_tag_meta=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='migrationtestmodel',
            name='tags',
            field=tagulous.models.fields.TagField(force_lowercase=True, to='tagulous_tests_migration.Tagulous_MigrationTestModel_tags', protect_initial=True, help_text=b'Enter a comma-separated tag string', case_sensitive=True, autocomplete_initial=True, autocomplete_view=b'tagulous_tests_app-null', autocomplete_limit=3, initial='Mr', protect_all=False, _set_tag_meta=True),
            preserve_default=True,
        ),
    ]



"""
Tagulous test: Model field order

Modules tested:
    tagulous.models.fields.SingleTagField
    tagulous.models.fields.TagField
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

from tests.lib import *


class ModelFieldOrderTest(TagTestManager, TestCase):
    """
    Test model SingleTagField order
    """
    manage_models = [
        test_models.MixedOrderTest,
    ]

    def test_order_correct(self):
        """
        Check that the order of the non-ManyToMany fields is correct
        This is to check that Django internals haven't changed significantly
        """
        # Check the ordering is as expected
        opts = test_models.MixedOrderTest._meta
        # Django 1.4 and 1.5 don't have concrete_fields in meta Options.
        # Django 1.8 states that it's an internal function and shouldn't be
        # used directly, but using it anyway to keep code simple.
        if hasattr(opts, 'concrete_fields'):
            concrete_fields = opts.concrete_fields
        else:
            concrete_fields = [f for f in opts.fields if f.column is not None]
        local_fields = sorted(concrete_fields + opts.many_to_many)
        expected_fields = [
            # Auto pk 'id'
            'id',
            # Defined fields
            'char1', 'fk1', 'char2', 'single1', 'char3', 'm2m1', 'char4',
            'multi1', 'char5', 'm2m2', 'char6', 'fk2', 'char7',
        ]
        self.assertEqual(len(local_fields), len(expected_fields))
        for i in range(len(local_fields)):
            self.assertEqual(local_fields[i].name, expected_fields[i])



from __future__ import unicode_literals

from django import forms
from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase
from django.http import HttpResponseRedirect
from django.shortcuts import render

from tagulous import models as tag_models
from tagulous import forms as tag_forms


###############################################################################
########################################################### Admin classes
###############################################################################

class TaggedBaseModelAdminMixin(admin.options.BaseModelAdmin):
    """
    Mixin for BaseModelAdmin subclasses which are for tagged models
    """
    def formfield_for_dbfield(self, db_field, **kwargs):
        """
        Remove the RelatedFieldWidgetWrapper from tag fields, so they don't
        display popup buttons
        """
        formfield = super(
            TaggedBaseModelAdminMixin, self
        ).formfield_for_dbfield(db_field, request=kwargs.pop('request', None), **kwargs)

        if (
            isinstance(db_field, (tag_models.SingleTagField, tag_models.TagField))
            and
            isinstance(formfield.widget, admin.widgets.RelatedFieldWidgetWrapper)
            and
            isinstance(formfield.widget.widget, tag_forms.AdminTagWidget)
        ):
            formfield.widget = formfield.widget.widget
        return formfield


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for Tagged models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class TaggedModelAdmin(TaggedBaseModelAdminMixin, admin.ModelAdmin):
    pass


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for TagModel models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class TagModelAdmin(TaggedBaseModelAdminMixin, admin.ModelAdmin):
    list_display = ['name', 'count', 'protected']
    list_filter = ['protected']
    exclude = ['count']
    actions = ['merge_tags']
    prepopulated_fields = {"slug": ("name",)}

    def merge_tags(self, request, queryset):
        """
        Admin action to merge tags
        """
        # Thanks to:
        #   http://www.hoboes.com/Mimsy/hacks/django-actions-their-own-intermediate-page/

        # Create a form
        is_tree = issubclass(self.model, tag_models.TagTreeModel)
        class MergeForm(forms.Form):
            # Keep selected items in same field, admin.ACTION_CHECKBOX_NAME
            _selected_action = forms.CharField(widget=forms.MultipleHiddenInput)
            # Allow use to select from selected items
            merge_to = forms.ModelChoiceField(queryset)

        if is_tree:
            class MergeForm(MergeForm):
                # Allow to merge recursively
                merge_children = forms.BooleanField(required=False)

        if 'merge' in request.POST:
            merge_form = MergeForm(request.POST)
            if merge_form.is_valid():
                # Merge - with children if set
                merge_to = merge_form.cleaned_data['merge_to']
                kwargs = {}
                if is_tree and merge_form.cleaned_data['merge_children']:
                    kwargs['children'] = True
                merge_to.merge_tags(queryset, **kwargs)

                # Django 1.4 doesn't support level=messages.SUCCESS
                self.message_user(request, 'Tags merged')
                return HttpResponseRedirect(request.get_full_path())

        else:
            tag_pks = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
            if len(tag_pks) < 2:
                # Django 1.4 doesn't support level=messages.SUCCESS
                self.message_user(
                    request, 'You must select at least two tags to merge',
                )
                return HttpResponseRedirect(request.get_full_path())

            merge_form = MergeForm(
                initial={
                    admin.ACTION_CHECKBOX_NAME: request.POST.getlist(
                        admin.ACTION_CHECKBOX_NAME
                    ),
                    'merge_children': True,
                }
            )

        return render(request, 'tagulous/admin/merge_tags.html', {
            'title': 'Merge tags',
            'opts': self.model._meta,
            'merge_form': merge_form,
            'tags': queryset,
        })
    merge_tags.short_description = 'Merge selected tags...'


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for TagTreeModel models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class TagTreeModelAdmin(TagModelAdmin):
    exclude = ['count', 'parent', 'path', 'label', 'level']


###############################################################################
########################################################### Admin registration
###############################################################################

# Give contrib.admin a default widget for tag fields
admin.options.FORMFIELD_FOR_DBFIELD_DEFAULTS.update({
    tag_models.SingleTagField:  {'widget': tag_forms.AdminTagWidget},
    tag_models.TagField:        {'widget': tag_forms.AdminTagWidget},
})

def _create_display(field):
    """
    ModelAdmin display function factory
    """
    def display(self, obj):
        return getattr(obj, field).get_tag_string()
    display.short_description = field.replace('_', ' ')
    return display


def enhance(model, admin_class):
    """
    Add tag support to the admin class based on the specified model
    """
    #
    # Get a list of all tag fields
    #

    # Dict of single tag fields, {name: tag}
    single_tag_fields = {}

    # Dict of normal tag fields, {name: tag}
    tag_fields = {}

    # List of all single and normal tag fields
    tag_field_names = []

    # Check for SingleTagField related fields
    for field in tag_models.singletagfields_from_model(model):
        single_tag_fields[field.name] = field
        tag_field_names.append(field.name)

    # Check for TagField m2m fields
    for field in tag_models.tagfields_from_model(model):
        tag_fields[field.name] = field
        tag_field_names.append(field.name)


    #
    # Ensure any tag fields in list_display are rendered by functions
    #

    # The admin.site.register will complain if it's a ManyToManyField, so this
    # will work around that.
    #
    # We also need to have a different name to the model field, otherwise the
    # ChangeList class will just use the model field - that would get the tag
    # strings showing in the table, but the column would be sortable which
    # would cause problems for TagFields, and the display function would never
    # get called, which would be unexpected for anyone maintaining this code.
    if hasattr(admin_class, 'list_display'):
        # Make sure we're working with a list
        if isinstance(admin_class.list_display, tuple):
            admin_class.list_display = list(admin_class.list_display)

        for i, field in enumerate(admin_class.list_display):
            # If the field's not a callable, and not in the admin class already
            if not hasattr(field, '__call__') and not hasattr(admin_class, field):
                # Only TagFields (admin can already handle SingleTagField FKs)
                if field in tag_fields:
                    # Create new field name and replace in list_display
                    display_name = '_tagulous_display_%s' % field
                    admin_class.list_display[i] = display_name

                    # Add display function to admin class
                    setattr(admin_class, display_name, _create_display(field))

    #
    # If admin is for a tag model, ensure any inlines for tagged models are
    # subclasses of TaggedInlineFormSet.
    #
    if (
        issubclass(model, tag_models.BaseTagModel)
        and hasattr(admin_class, 'inlines')
    ):
        for inline_cls in admin_class.inlines:
            # Make sure inline class uses TaggedBaseModelAdminMixin
            if not issubclass(inline_cls, TaggedBaseModelAdminMixin):
                inline_cls.__bases__ = (
                    TaggedBaseModelAdminMixin,
                ) + inline_cls.__bases__

            # Make sure inlines used TaggedInlineFormSet
            if (
                issubclass(inline_cls.model, tag_models.TaggedModel)
                and not
                issubclass(inline_cls.formset, tag_forms.TaggedInlineFormSet)
            ):
                orig_cls = inline_cls.formset
                inline_cls.formset = type(
                    str('Tagged%s' % orig_cls.__name__),
                    (tag_forms.TaggedInlineFormSet, orig_cls),
                    {},
                )


def register(model, admin_class=None, site=None, **options):
    """
    Provide tag support to the model when it is registered with the admin site.

    For tagged models (have one or more SingleTagField or TagField fields):
        * Admin will support TagField in list_display

    For tag models (subclass of TagModel):
        * Admin will provide a merge action to merge tags

    For other models:
        * No changes made

    Arguments:
        model       Model to register
        admin_class Admin class for model
        site        Admin site to register with
                    Default: django.contrib.admin.site
        **options   Extra options for admin class

    This only supports one model, but is otherwise safe to use with non-tagged
    models.
    """
    # Look at the model we've been given
    if isinstance(model, tag_models.BaseTagDescriptor):
        # It's a tag descriptor; change it for the tag model itself
        model = model.tag_model

    elif not isinstance(model, ModelBase):
        raise ImproperlyConfigured(
            'Tagulous can only register a single model with admin.'
        )

    # Ensure we have a valid admin site
    if site is None:
        site = admin.site

    #
    # Determine appropriate admin class
    #
    if not admin_class:
        admin_class = admin.ModelAdmin

    # Going to make a list of base classes to inject
    cls_bases = []

    # If it's a tag model, subclass TagModelAdmin or TagTreeModelAdmin
    if issubclass(model, tag_models.BaseTagModel):
        if issubclass(model, tag_models.TagTreeModel):
            if not issubclass(admin_class, TagTreeModelAdmin):
                cls_bases += [TagTreeModelAdmin]
        else:
            if not issubclass(admin_class, TagModelAdmin):
                cls_bases += [TagModelAdmin]

    # If it's a tagged model, subclass TaggedModelAdmin
    singletagfields = tag_models.singletagfields_from_model(model)
    tagfields = tag_models.tagfields_from_model(model)
    if singletagfields or tagfields:
        if not issubclass(admin_class, TaggedModelAdmin):
            cls_bases += [TaggedModelAdmin]

    # If options specified, or other bases, will need to subclass admin_class
    if options or cls_bases:
        cls_bases += [admin_class]
        # Update options with anything the new subclasses could have overidden
        # in a custom ModelAdmin - unless they're already overridden in options
        options['__module__'] = __name__
        if admin_class != admin.ModelAdmin:
            options.update(dict(
                (k, v) for k, v in admin_class.__dict__.items()
                if k in ['list_display', 'list_filter', 'exclude', 'actions']
                and k not in options
            ))
        admin_class = type(
            str("%sAdmin" % model.__name__), tuple(cls_bases), options,
        )

    # Enhance the model admin class
    enhance(model, admin_class)

    # Register the model
    # Don't pass options - we've already dealt with that
    site.register(model, admin_class)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Removing unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug']
        db.delete_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug'])
        # Adding field 'Tagulous_MigrationTestModel_tags.parent'
        db.add_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'parent',
                      self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='children', null=True, to=orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']),
                      keep_default=False)

        # Adding field 'Tagulous_MigrationTestModel_tags.label'
        db.add_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'label',
                      self.gf('django.db.models.fields.CharField')(default='.', max_length=191),
                      keep_default=False)

        # Adding field 'Tagulous_MigrationTestModel_tags.level'
        db.add_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'level',
                      self.gf('django.db.models.fields.IntegerField')(default=1),
                      keep_default=False)

        # Adding field 'Tagulous_MigrationTestModel_tags.path'
        from tagulous.models.migrations import add_unique_column
        add_unique_column(
            self, db,
            orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags'],
            'path',
            lambda obj: setattr(obj, 'path', str(obj.pk)),
            'django.db.models.fields.TextField',
        )

        # Adding unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug', 'parent']
        db.create_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug', 'parent_id'])


    def backwards(self, orm):
        # Removing unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug', 'parent']
        db.delete_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug', 'parent_id'])

        # Deleting field 'Tagulous_MigrationTestModel_tags.parent'
        db.delete_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'parent_id')

        # Deleting field 'Tagulous_MigrationTestModel_tags.path'
        db.delete_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'path')

        # Deleting field 'BookmarkTag.label'
        db.delete_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'label')

        # Deleting field 'BookmarkTag.level'
        db.delete_column(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', 'level')

        # Adding unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug']
        db.create_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug'])


    models = {
        u'tagulous_tests_migration.tagulous_migrationtestmodel_singletag': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug',),)", 'object_name': 'Tagulous_MigrationTestModel_singletag', '_bases': ['tagulous.models.BaseTagModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.tagulous_migrationtestmodel_tags': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug', 'parent'),)", 'object_name': 'Tagulous_MigrationTestModel_tags', '_bases': ['tagulous.models.BaseTagTreeModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'label': ('django.db.models.fields.CharField', [], {'max_length': '191'}),
            'level': ('django.db.models.fields.IntegerField', [], {'default': '1'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']"}),
            'path': ('django.db.models.fields.TextField', [], {'unique': 'True'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.migrationtestmodel': {
            'Meta': {'object_name': 'MigrationTestModel'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'singletag': ('tagulous.models.fields.SingleTagField', [], {'_set_tag_meta': 'True', 'blank': 'True', 'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_singletag']", 'null': 'True'}),
            'tags': ('tagulous.models.fields.TagField', [], {'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']", 'tree': 'True', '_set_tag_meta': 'True'})
        }
    }

    complete_apps = ['tagulous_tests_migration']


# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Tagulous_MigrationTestModel_tags'
        db.create_table(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=191)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=50)),
            ('count', self.gf('django.db.models.fields.IntegerField')(default=0)),
            ('protected', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'tagulous_tests_migration', ['Tagulous_MigrationTestModel_tags'])

        # Adding unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug']
        db.create_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug'])

        # Adding model 'Tagulous_MigrationTestModel_singletag'
        db.create_table(u'tagulous_tests_migration_tagulous_migrationtestmodel_singletag', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=191)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=50)),
            ('count', self.gf('django.db.models.fields.IntegerField')(default=0)),
            ('protected', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'tagulous_tests_migration', ['Tagulous_MigrationTestModel_singletag'])

        # Adding unique constraint on 'Tagulous_MigrationTestModel_singletag', fields ['slug']
        db.create_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_singletag', ['slug'])

        # Adding field 'MigrationTestModel.singletag'
        db.add_column(u'tagulous_tests_migration_migrationtestmodel', 'singletag',
                      self.gf('tagulous.models.fields.SingleTagField')(_set_tag_meta=True, null=True, to=orm['tagulous_tests_migration.Tagulous_MigrationTestModel_singletag'], blank=True),
                      keep_default=False)

        # Adding M2M table for field tags on 'MigrationTestModel'
        m2m_table_name = db.shorten_name(u'tagulous_tests_migration_migrationtestmodel_tags')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('migrationtestmodel', models.ForeignKey(orm[u'tagulous_tests_migration.migrationtestmodel'], null=False)),
            ('tagulous_migrationtestmodel_tags', models.ForeignKey(orm[u'tagulous_tests_migration.tagulous_migrationtestmodel_tags'], null=False))
        ))
        db.create_unique(m2m_table_name, ['migrationtestmodel_id', 'tagulous_migrationtestmodel_tags_id'])


    def backwards(self, orm):
        # Removing unique constraint on 'Tagulous_MigrationTestModel_singletag', fields ['slug']
        db.delete_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_singletag', ['slug'])

        # Removing unique constraint on 'Tagulous_MigrationTestModel_tags', fields ['slug']
        db.delete_unique(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags', ['slug'])

        # Deleting model 'Tagulous_MigrationTestModel_tags'
        db.delete_table(u'tagulous_tests_migration_tagulous_migrationtestmodel_tags')

        # Deleting model 'Tagulous_MigrationTestModel_singletag'
        db.delete_table(u'tagulous_tests_migration_tagulous_migrationtestmodel_singletag')

        # Deleting field 'MigrationTestModel.singletag'
        db.delete_column(u'tagulous_tests_migration_migrationtestmodel', 'singletag_id')

        # Removing M2M table for field tags on 'MigrationTestModel'
        db.delete_table(db.shorten_name(u'tagulous_tests_migration_migrationtestmodel_tags'))


    models = {
        u'tagulous_tests_migration.tagulous_migrationtestmodel_singletag': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug',),)", 'object_name': 'Tagulous_MigrationTestModel_singletag', '_bases': ['tagulous.models.BaseTagModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.tagulous_migrationtestmodel_tags': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('slug',),)", 'object_name': 'Tagulous_MigrationTestModel_tags', '_bases': ['tagulous.models.BaseTagModel']},
            'count': ('django.db.models.fields.IntegerField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '191'}),
            'protected': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '50'})
        },
        u'tagulous_tests_migration.migrationtestmodel': {
            'Meta': {'object_name': 'MigrationTestModel'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            'singletag': ('tagulous.models.fields.SingleTagField', [], {'_set_tag_meta': 'True', 'null': 'True', 'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_singletag']", 'blank': 'True'}),
            'tags': ('tagulous.models.fields.TagField', [], {'to': u"orm['tagulous_tests_migration.Tagulous_MigrationTestModel_tags']", '_set_tag_meta': 'True'})
        }
    }

    complete_apps = ['tagulous_tests_migration']


"""
Python serializer with Tagulous support
"""
from __future__ import unicode_literals

from django.core.serializers import python as python_serializer
from django.utils import six

from tagulous.serializers import base


class Serializer(base.SerializerMixin, python_serializer.Serializer):
    """
    Serializes a QuerySet to basic Python objects, with tag field support
    """
    pass


Deserializer = base.DeserializerWrapper(python_serializer.Deserializer, doc=(
    "Deserialize Python objects into Django ORM instances, with tag field "
    "support"
))

base.monkeypatch_get_model(python_serializer)



"""
Tagulous test: south migrations

Modules tested:
    tagulous.models.migrations
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import inspect
import os
import sys
import shutil
import warnings

from django.conf import settings
from django.core.management import call_command
from django.db import DatabaseError
from django.utils import six

from tests.lib import *
from tests import tagulous_tests_migration

try:
    import south
except ImportError:
    south = None

try:
    eval("u'string'")
except SyntaxError:
    # South doesn't support Python 3.2
    south = None

# If True, display output from call_command - use for debugging tests
DISPLAY_CALL_COMMAND = False

app_name = 'tagulous_tests_migration'
app_module = sys.modules['tests.%s' % app_name]
migrations_name = 'migrations_%s' % testenv
migrations_module = 'tests.%s.%s' % (app_name, migrations_name)
migrations_path = None


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Util functions
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

def clear_migrations():
    "Clear cached mentions of south migrations to force a reload"
    from south.migration import Migrations

    # Clear metaclass cache
    if app_name in Migrations.instances:
        del Migrations.instances[app_name]

    # Remove loaded migrations
    if hasattr(app_module, migrations_name):
        delattr(app_module, migrations_name)

    for key in list(sys.modules.keys()):
        if key.startswith(migrations_module):
            del sys.modules[key]

def get_migrations():
    "Clears migration cache and gets migrations for the test app"
    from south.migration import Migrations
    clear_migrations()
    return Migrations(
        tagulous_tests_migration, force_creation=True, verbose_creation=False,
    )

def get_migrations_dir():
    "Get migration dir"
    if migrations_path is None:
        globals()['migrations_path'] = get_migrations().migrations_dir()
    return migrations_path

def get_expected_dir():
    return os.path.normpath(
        os.path.join(
            get_migrations_dir(),
            '..',
            'south_migrations_expected',
        ),
    )

def clean_all():
    "Clean everything - roll back and delete any migrations, forget any loaded"
    # Delete migrations
    migrations_dir = get_migrations_dir()
    expected_dir = get_expected_dir()
    if not (
        app_name in migrations_dir
        and migrations_dir.endswith(migrations_name)
    ):
        # Catch unexpected path - don't want to delete anything important
        raise ValueError('Migrations dir has unexpected name')
    if os.path.isdir(migrations_dir):
        shutil.rmtree(migrations_dir)
    elif os.path.exists(migrations_dir):
        raise ValueError('Migrations dir is not a dir')

    # Try to roll back to zero using expected migrations
    shutil.copytree(expected_dir, migrations_dir)

    try:
        migrate_app(target='zero')
    except DatabaseError:
        # Guess it didn't exist - that's ok, nothing to reverse
        pass
    shutil.rmtree(migrations_dir)

    # Empty models
    tagulous_tests_migration.models.unset_model()

    # Clear south's migration cache
    clear_migrations()


def migrate_app(target=None):
    "Apply migrations"
    clear_migrations()

    if DISPLAY_CALL_COMMAND:
        print(">> manage.py migrate %s target=%s" % (app_name, target))

    try:
        with Capturing() as output:
            with warnings.catch_warnings(record=True) as cw:
                call_command(
                    'migrate',
                    app_name,       # app to migrate
                    target=target,  # Optional target
                    verbosity=1,    # Silent
                )
    except Exception as e:
        print(">> Migration failed:")
        print("\n".join(output))
        print("<<<<<<<<<<")
        raise e

    # Ensure caught warnings are expected
    if django.VERSION < (1, 6):
        assert len(cw) == 0
    else:
        for w in cw:
            assert issubclass(w.category, PendingDeprecationWarning)

    if DISPLAY_CALL_COMMAND:
        print('\n'.join(output))
        print("<<<<<<<<<<")

    return output



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Tests
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

@unittest.skipIf(django.VERSION >= (1, 7), 'South tests not run for Django 1.7+')
@unittest.skipIf(south is None, 'South not installed')
@unittest.skipIf(
    'south' not in settings.INSTALLED_APPS, 'South not in INSTALLED_APPS'
)
class SouthMigrationTest(TagTestManager, TransactionTestCase):
    """
    Test south migrations
    """
    #
    # Test management - ensure it's clean before and after
    #

    @classmethod
    def setUpClass(self):
        "Clean everything before each test in case previous run failed"
        clean_all()

    def tearDownExtra(self):
        "Clean away the model so it's not installed by TestCase"
        clean_all()

    @classmethod
    def tearDownClass(cls):
        "Leave everything clean at the end of the tests"
        clean_all()


    #
    # Migration file analysis
    #

    def _import_migration(self, path, name):
        "Import the named migration from the given file path"
        root = os.path.dirname(__file__)
        if not path.startswith(root):
            return self.fail(
                'Could not find common root between test and migration dir:\n'
                '  %s\n  %s' % __file__, path
            )

        module_name = '.'.join(
            ['tests'] + path[len(root) + 1:].split(os.sep) + [name]
        )
        module = __import__(module_name, {}, {}, ['Migration'])

        return module

    def _parse_migration_function(self, fn):
        """
        Parse a migration function into a dict

        Dict will have keys for each db method called, with values as dicts of
        whatever makes them unique and the arguments.

        There will also be a 'vars' dict with variable definitions
        """
        import ast
        lines, firstlineno = inspect.getsourcelines(fn.__code__)

        # De-indent and parse into an AST
        indent = len(lines[0]) - len(lines[0].lstrip())
        lines = [line[indent:] for line in lines]
        root = ast.parse(''.join(lines))

        def ast_dump(*nodes):
            "Dump ast node without unicode strings"
            return ', '.join(
                ast.dump(node).replace(
                    "Str(s=u'", "Str(s='"
                ).replace(
                    "{u'", "{'"
                )
                for node in nodes
            )

        def parse_field(field):
            "Parse a field into a tuple of (cls_name, 'ast value')"
            if isinstance(field.func, ast.Attribute):
                # Call to self.gf
                field_cls = six.text_type(
                    field.func.value.id + '.' + field.func.attr
                )
            else:
                # Direct ref to field
                field_cls = six.text_type(field.func.args[0].s)

            attrs = {}
            for keyword in field.keywords:
                attrs[keyword.arg] = ast_dump(keyword.value)
            return (field_cls, attrs)

        # Find all db object method calls
        data = {
            'create_table':         {},
            'send_create_signal':   {},
            'create_unique':        {},
            'add_column':           {},
            'shorten_name':         {},
            'delete_table':         {},
        }
        ast_state = {}

        # This seemed like such a good idea at the time
        for node in root.body[0].body:
            # db.create_table( <arguments> )
            dbcall = node.value
            self.assertEqual(dbcall.func.value.id, 'db')
            method = dbcall.func.attr

            # Shift arguments node tree into dict
            if method == 'create_table':
                # <table>, <fields tuple>
                if isinstance(dbcall.args[0], ast.Str):
                    table_name = six.text_type(dbcall.args[0].s)
                elif isinstance(dbcall.args[0], ast.Name):
                    var_name = dbcall.args[0].id
                    self.assertTrue(var_name in ast_state)
                    table_name = ast_state[var_name]
                self.assertFalse(table_name in data[method])

                fields = {}
                for field_tuple in dbcall.args[1].elts:
                    # <field name>, self.gf(<type>)(<args>)
                    field_name = six.text_type(field_tuple.elts[0].s)
                    self.assertFalse(field_name in fields)
                    fields[field_name] = parse_field(field_tuple.elts[1])
                data[method][table_name] = fields

            elif method == 'send_create_signal':
                # <app name>, [<model name>]
                model_name = six.text_type(dbcall.args[1].elts[0].s)
                self.assertFalse(model_name in data[method])
                data[method][model_name] = ast_dump(dbcall)

            elif method == 'create_unique':
                # <table name>, [<field name>]
                if isinstance(dbcall.args[0], ast.Str):
                    table_name = six.text_type(dbcall.args[0].s)
                elif isinstance(dbcall.args[0], ast.Name):
                    var_name = dbcall.args[0].id
                    self.assertTrue(var_name in ast_state)
                    table_name = ast_state[var_name]
                self.assertFalse(table_name in data[method])
                data[method][table_name] = ast_dump(dbcall.args[1])

            elif method == 'add_column':
                # <table name>, <field name>, self.gf(<type>)(<args>), keep_default=<?>
                field_id = six.text_type(
                    dbcall.args[0].s) + '.' + str(dbcall.args[1].s
                )
                self.assertFalse(field_id in data[method])
                data[method][field_id] = [
                    parse_field(dbcall.args[2]),
                    ast_dump(*dbcall.args[3:]),
                ]

            elif method == 'shorten_name':
                # <var> = db.shorten_name(<table name>)
                table_name = six.text_type(dbcall.args[0].s)
                self.assertFalse(field_id in data[method])
                data[method][table_name] = ast_dump(node.targets[0])
                ast_state[node.targets[0].id] = table_name

            else:
                self.fail('Unrecognised db method: %s' % method)

        return data


    #
    # Extra assertions
    #

    def assertValidMigration(self, module):
        "Basic validation of an imported migration"
        self.assertTrue('Migration' in module.__dict__)
        self.assertTrue(issubclass(module.Migration, south.v2.SchemaMigration))
        self.assertTrue(hasattr(module.Migration, 'forwards'))
        self.assertTrue(hasattr(module.Migration, 'backwards'))
        self.assertTrue(hasattr(module.Migration, 'models'))

    def assertMigrationExpected(self, name):
        "Compare two migration files"
        # Import and validate the migrations
        mig1 = self._import_migration(get_migrations_dir(), name)
        self.assertValidMigration(mig1)

        mig2 = self._import_migration(get_expected_dir(), name)
        self.assertValidMigration(mig2)

        # Fn to standardise models dict string type
        def fix_text_type(val):
            if isinstance(val, six.string_types):
                val = val.replace("u'", "'")
                if six.PY3:
                    return bytes(val, 'utf8')
                return bytes(val)
            elif isinstance(val, dict):
                dct = val
                for key, val in dct.items():
                    del dct[key]
                    dct[fix_text_type(key)] = fix_text_type(val)
                return dct
            else:
                # Assume it's an iterable
                val = [fix_text_type(v) for v in val]

        # Compare models - some strings will be unicode, some not
        self.assertEqual(
            fix_text_type(mig1.Migration.models),
            fix_text_type(mig2.Migration.models),
        )

        data1 = self._parse_migration_function(mig1.Migration.forwards)
        data2 = self._parse_migration_function(mig2.Migration.forwards)
        self.assertEqual(data1, data2)


    #
    # Tests
    #

    def migrate_initial(self):
        """
        Load the initial test app's model, and create and apply an initial
        schema migration. Check that the migration worked.
        """
        model_initial = tagulous_tests_migration.models.set_model_initial()
        self.assertFalse(
            issubclass(model_initial, tag_models.tagged.TaggedModel)
        )

        # Ensure migration dir exists - we're using a custom one, so South
        # won't create it
        migrations_dir = get_migrations_dir()
        os.mkdir(migrations_dir)
        open(os.path.join(migrations_dir, '__init__.py'), 'a').close()

        # Run schemamigration --initial
        with Capturing() as output:
            call_command(
                'schemamigration',
                # First two args not named - kwarg 'name' clashes
                app_name,       # app to migrate
                'initial',      # name - Name of migration to create
                initial=True,   # This is the initial schema migration
                verbosity=0,    # Silent
            )

        # Check the files were created as expected
        migrations = get_migrations()
        migrations = [six.text_type(m) for m in migrations]
        self.assertEqual(len(migrations), 1)
        self.assertEqual(migrations[0], '%s:0001_initial' % app_name)
        self.assertMigrationExpected('0001_initial')

        # Check they apply correctly
        output = migrate_app()
        self.assertSequenceEqual(output, [
            'Running migrations for tagulous_tests_migration:',
            ' - Migrating forwards to 0001_initial.',
            ' > tagulous_tests_migration:0001_initial',
            ' - Loading initial data for tagulous_tests_migration.',
            'Installed 0 object(s) from 0 fixture(s)'
        ])

        return model_initial

    def migrate_tagged(self):
        """
        After migrating to the initial model, switch the test app to the
        tagged model, and create and apply a schema migration. Check that the
        migration worked.
        """
        # First need to migrate to initial - re-run that migration
        self.migrate_initial()

        # Now switch model
        model_tagged = tagulous_tests_migration.models.set_model_tagged()
        self.assertTrue(issubclass(model_tagged, tag_models.tagged.TaggedModel))

        # Run schemamigration --auto
        with Capturing() as output:
            call_command(
                'schemamigration',
                app_name,       # app to migrate
                'tagged',       # name - Name of migration to create
                auto=True,      # This is an auto schema migration
                verbosity=0,    # Silent
            )

        # Check the files were created as expected
        migrations = get_migrations()
        migrations = [six.text_type(m) for m in migrations]
        self.assertEqual(len(migrations), 2)
        self.assertEqual(migrations[0], '%s:0001_initial' % app_name)
        self.assertEqual(migrations[1], '%s:0002_tagged' % app_name)
        self.assertMigrationExpected('0002_tagged')

        # Check they apply correctly
        output = migrate_app()
        '''
        self.assertSequenceEqual(output, [
            'Running migrations for tagulous_tests_migration:',
            ' - Migrating forwards to 0002_tagged.',
            ' > tagulous_tests_migration:0002_tagged',
            ' - Loading initial data for tagulous_tests_migration.',
            'Installed 0 object(s) from 0 fixture(s)'
        ])
        '''

        return model_tagged

    def migrate_tree(self):
        """
        After migrating to the tagged model, switch the test app to the tree
        model, load some test data, and apply the pre-written schema migration
        which uses add_unique_column. Check that the migration worked.
        """
        # Migrate to tagged
        model_tagged = self.migrate_tagged()

        # Add some test data
        model_tagged.tags.tag_model.objects.create(name='one/two/three')
        model_tagged.tags.tag_model.objects.create(name='uno/dos/tres')
        self.assertTagModel(model_tagged.tags.tag_model, {
            'one/two/three': 0,
            'uno/dos/tres':  0,
        })

        # Now switch model
        model_tree = tagulous_tests_migration.models.set_model_tree()
        self.assertTrue(issubclass(model_tree, tag_models.tagged.TaggedModel))
        self.assertTagModel(model_tagged.tags.tag_model, {
            'one/two/three': 0,
            'uno/dos/tres':  0,
        })

        # We can't create a schema migration here; because we're adding a null
        # field, South would ask us questions we don't want to answer anyway,
        # because we'd replace the add_column call with a call to
        # tagulous.models.migration.add_unique_column. We'll therefore use one
        # we prepared earlier, 0003_tree.py
        #
        # But first, confirm schemamigration would have correctly detected the
        # tag model base has changed to a BaseTagTreeModel:
        frozen_singletag = south.creator.freezer.prep_for_freeze(
            model_tree.singletag.tag_model
        )
        self.assertSequenceEqual(
            frozen_singletag['Meta']['_bases'],
            ['tagulous.models.BaseTagModel'],
        )

        frozen_tags = south.creator.freezer.prep_for_freeze(
            model_tree.tags.tag_model
        )
        self.assertSequenceEqual(
            frozen_tags['Meta']['_bases'],
            ['tagulous.models.BaseTagTreeModel'],
        )

        # Add in the prepared schemamigration for the tree
        migrations_dir = get_migrations_dir()
        expected_dir = get_expected_dir()
        shutil.copy(
            os.path.join(expected_dir, '0003_tree.py'),
            migrations_dir
        )

        # Check the files were created as expected
        migrations = get_migrations()
        migrations = [six.text_type(m) for m in migrations]
        self.assertEqual(len(migrations), 3)
        self.assertEqual(migrations[0], '%s:0001_initial' % app_name)
        self.assertEqual(migrations[1], '%s:0002_tagged' % app_name)
        self.assertEqual(migrations[2], '%s:0003_tree' % app_name)

        # Check they apply correctly
        output = migrate_app()
        '''
        self.assertSequenceEqual(output, [
            'Running migrations for tagulous_tests_migration:',
            ' - Migrating forwards to 0003_tree.',
            ' > tagulous_tests_migration:0003_tree',
            ' - Loading initial data for tagulous_tests_migration.',
            'Installed 0 object(s) from 0 fixture(s)'
        ])
        '''

        # Data shouldn't have changed yet
        self.assertTagModel(model_tree.tags.tag_model, {
            'one/two/three':    0,
            'uno/dos/tres':     0,
        })
        self.assertEqual(model_tree.tags.tag_model.objects.get(pk=1).path, '1')
        self.assertEqual(model_tree.tags.tag_model.objects.get(pk=2).path, '2')

        # Rebuild tree
        model_tree.tags.tag_model.objects.rebuild()

        # We should now have nicely-built trees
        self.assertTagModel(model_tree.tags.tag_model, {
            'one':              0,
            'one/two':          0,
            'one/two/three':    0,
            'uno':              0,
            'uno/dos':          0,
            'uno/dos/tres':     0,
        })
        tag_objects = model_tree.tags.tag_model.objects
        self.assertEqual(tag_objects.get(name='one').path, 'one')
        self.assertEqual(tag_objects.get(name='one/two').path, 'one/two')
        self.assertEqual(tag_objects.get(name='one/two/three').path, 'one/two/three')
        self.assertEqual(tag_objects.get(name='uno').path, 'uno')
        self.assertEqual(tag_objects.get(name='uno/dos').path, 'uno/dos')
        self.assertEqual(tag_objects.get(name='uno/dos/tres').path, 'uno/dos/tres')

        return model_tree

    def migrate_data(self):
        """
        After migrating to the tree model, apply the pre-written data migration
        which tests tag fields and models. Check that the migration worked.
        """
        model_tree = self.migrate_tree()

        # Empty the tags from the test model added by migrate_tree
        model_tree.tags.tag_model.objects.all().delete()
        self.assertTagModel(model_tree.tags.tag_model, {})

        # Add some test data to the model itself
        model_tree.objects.create(
            name='Test 1', singletag='Mr', tags='one/two, uno/dos',
        )
        model_tree.objects.create(
            name='Test 2', singletag='Mrs', tags='one/two',
        )
        model_tree.objects.create(
            name='Test 3', singletag='Mr', tags='uno/dos',
        )
        self.assertTagModel(model_tree.singletag.tag_model, {
            'Mr':       2,
            'Mrs':      1,
        })
        self.assertTagModel(model_tree.tags.tag_model, {
            'one':      0,
            'one/two':  2,
            'uno':      0,
            'uno/dos':  2,
        })

        # Add in the datamigration
        migrations_dir = get_migrations_dir()
        expected_dir = get_expected_dir()
        shutil.copy(
            os.path.join(expected_dir, '0004_data.py'),
            migrations_dir
        )

        # Check the files were created as expected
        migrations = get_migrations()
        migrations = [six.text_type(m) for m in migrations]
        self.assertEqual(len(migrations), 4)
        self.assertEqual(migrations[0], '%s:0001_initial' % app_name)
        self.assertEqual(migrations[1], '%s:0002_tagged' % app_name)
        self.assertEqual(migrations[2], '%s:0003_tree' % app_name)
        self.assertEqual(migrations[3], '%s:0004_data' % app_name)

        # Check they apply correctly
        output = migrate_app()
        self.assertSequenceEqual(output, [
            'Running migrations for tagulous_tests_migration:',
            ' - Migrating forwards to 0004_data.',
            ' > tagulous_tests_migration:0004_data',
            " - Migration 'tagulous_tests_migration:0004_data' is "
                "marked for no-dry-run.",
            ' - Loading initial data for tagulous_tests_migration.',
            'Installed 0 object(s) from 0 fixture(s)'
        ])



    #
    # Tests
    #

    # Individual tests for development purposes
    # No point running each of them - test_data() will run them as its setup
    '''
    def test_initial(self):
        "Test initial migration is created and can be applied and used"
        self.migrate_initial()

    def test_tagged(self):
        "Test tagged migration is created and can be applied and used"
        self.migrate_tagged()

    def test_tree(self):
        "Test migration to Tree model using add_unique_column"
        self.migrate_tree()
    '''

    def test_data(self):
        "Test data migration"
        self.migrate_data()
